﻿using System;
using System.Threading;
using System.Collections.Generic;

using wclCommon;
using wclBluetooth;

namespace BluetoothManager
{
    class Program
    {
        private static wclBluetoothRadio Radio = null;
        private static AutoResetEvent CompleteEvent = null;
        private static ManualResetEvent TermintaeEvent = null;
        private static ManualResetEvent InitEvent = null;
        private static Int32 InitResult = wclErrors.WCL_E_SUCCESS;
        private static Thread RunThread = null;
        private static wclBluetoothManager Manager = null;
        private static Object Cs = new Object();

        private static void WriteLn(String s)
        {
            lock (Cs)
            {
                Console.WriteLine(s);
            }
        }

        private static void GetRadioInformation()
        {
            Console.Clear();

            WriteLn("RADIO INFORMATION");
            WriteLn("=================");
            WriteLn("  Driver: " + Radio.ApiName);
            WriteLn("  Available: " + Radio.Available.ToString());

            Boolean BoolValue;
            Int32 Res = Radio.GetConnectable(out BoolValue);
            if (Res != wclErrors.WCL_E_SUCCESS)
                WriteLn("  GetConnectable: error 0x" + Res.ToString("X8"));
            else
                WriteLn("  GetConnectable: " + BoolValue.ToString());

            Res = Radio.GetDiscoverable(out BoolValue);
            if (Res != wclErrors.WCL_E_SUCCESS)
                WriteLn("  Discoverable: error 0x" + Res.ToString("X8"));
            else
                WriteLn("  Discoverable: " + BoolValue.ToString());

            Byte Version;
            UInt16 Revision;
            Res = Radio.GetHciVersion(out Version, out Revision);
            if (Res != wclErrors.WCL_E_SUCCESS)
                WriteLn("  HCI version: error 0x" + Res.ToString("X8"));
            else
                WriteLn("  HCI version: " + Version.ToString() + "." + Revision.ToString());

            Res = Radio.GetLmpVersion(out Version, out Revision);
            if (Res != wclErrors.WCL_E_SUCCESS)
                WriteLn("  LMP version: error 0x" + Res.ToString("X8"));
            else
                WriteLn("  LMP version: " + Version.ToString() + "." + Revision.ToString());

            UInt16 Manu;
            Res = Radio.GetManufacturer(out Manu);
            if (Res != wclErrors.WCL_E_SUCCESS)
                WriteLn("  Manufacturer: error 0x" + Res.ToString("X8"));
            else
                WriteLn("  Manufacturer: " + Manu.ToString("X4"));

            String Name;
            Res = Radio.GetName(out Name);
            if (Res != wclErrors.WCL_E_SUCCESS)
                WriteLn("  Name: error 0x" + Res.ToString("X8"));
            else
                WriteLn("  Name: " + Name);

            WriteLn("Press any key to continue");
            Console.ReadKey();
        }

        private static void SetDiscoverable(Boolean On)
        {
            Console.Clear();
            WriteLn("CHANGE DISCOVERABLE MODE");
            WriteLn("========================");
            Int32 Res = Radio.SetDiscoverable(On);
            if (Res != wclErrors.WCL_E_SUCCESS)
                WriteLn("  Change discoverable mode failed: 0x" + Res.ToString("X8"));
            else
                WriteLn("  Discoverable mode has been changed");
            WriteLn("Press any key to continue");
            Console.ReadKey();
        }

        private static void SetConnectable(Boolean On)
        {
            Console.Clear();
            WriteLn("CHANGE CONNECTABLE MODE");
            WriteLn("=======================");
            Int32 Res = Radio.SetConnectable(On);
            if (Res != wclErrors.WCL_E_SUCCESS)
                WriteLn("  Change connectable mode failed: 0x" + Res.ToString("X8"));
            else
                WriteLn("  Connectable mode has been changed");
            WriteLn("Press any key to continue");
            Console.ReadKey();
        }

        private static void SwitchRadio(Boolean On)
        {
            Console.Clear();
            WriteLn("CHANGE RADIO STATE");
            WriteLn("===================");
            Int32 Res;
            if (On)
                Res = Radio.TurnOn();
            else
                Res = Radio.TurnOff();
            if (Res != wclErrors.WCL_E_SUCCESS)
                WriteLn("  Change radio state failed: 0x" + Res.ToString("X8"));
            else
                WriteLn("  Radio state has been changed");
            WriteLn("Press any key to continue");
            Console.ReadKey();
        }

        private static void Discover(Boolean Ble)
        {
            Console.Clear();
            WriteLn("DISOCVER REMOTE DEVICES");
            WriteLn("=======================");

            wclBluetoothDiscoverKind Kind;
            if (Ble)
                Kind = wclBluetoothDiscoverKind.dkBle;
            else
                Kind = wclBluetoothDiscoverKind.dkClassic;
            Int32 Res = Radio.Discover(10, Kind);
            if (Res != wclErrors.WCL_E_SUCCESS)
                WriteLn("Failed to start discovering: 0x" + Res.ToString("X8"));
            else
                CompleteEvent.WaitOne();

            WriteLn("Press any key to continue");
            Console.ReadKey();
        }

        private static void EnumConnected()
        {
            while (true)
            {
                Console.Clear();
                WriteLn("ENUMARATE CONNECTED DEVICES");
                WriteLn("===========================");

                Int64[] Devices;
                Int32 Res = Radio.EnumConnectedDevices(out Devices);
                if (Res != wclErrors.WCL_E_SUCCESS || Devices == null || Devices.Length == 0)
                {
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        WriteLn("Failed to enumerate connected devices: 0x" + Res.ToString("X8"));
                    else
                        WriteLn("No connected devices found");
                    WriteLn("Press any key to continue");
                    Console.ReadKey();
                    return;
                }

                WriteLn("Select device to disconnect");
                Byte i = 0;
                foreach (Int64 d in Devices)
                {
                    i++;
                    WriteLn("  " + i.ToString() + " : " + d.ToString("X12"));
                }
                WriteLn("  0 - Exit");

                WriteLn("Select option: ");
                String opt;
                while (true)
                {
                    opt = Console.ReadLine();
                    if (opt == "0")
                        return;

                    if (Convert.ToByte(opt) > i)
                    {
                        WriteLn("Invalid device number");
                        continue;
                    }

                    Res = Radio.RemoteDisconnect(Devices[i - 1]);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                    {
                        WriteLn("Disconnect failed: 0x" + Res.ToString("X8"));
                        continue;
                    }

                    WriteLn("Disconnected");
                    WriteLn("Press any key to continue");
                    Console.ReadKey();
                    break;
                }
            }
        }

        private static void Manager_OnDiscoveringStarted(Object Sender, wclBluetoothRadio Radio)
        {
            WriteLn("Discovering started");
        }

        private static void Manager_OnDiscoveringCompleted(Object Sender, wclBluetoothRadio Radio, Int32 Error)
        {
            WriteLn("Discovering completed with result: 0x" + Error.ToString("X8"));
            CompleteEvent.Set();
        }

        private static void Manager_OnDeviceFound(Object Sender, wclBluetoothRadio Radio, Int64 Address)
        {
            WriteLn("Device found: " + Address.ToString("X12"));
        }

        private static void Manager_OnStatusChanged(Object Sender, wclBluetoothRadio Radio)
        {
            WriteLn("  == Radio " + Radio.ApiName + " status changed. Available: " + Radio.Available.ToString() + " ==");
        }

        private static void Manager_AfterOpen(Object sender, EventArgs e)
        {
            WriteLn("Bluetooth manager opened");
        }

        private static void ManagerThread()
        {
            Manager = new wclBluetoothManager();
            Manager.AfterOpen += Manager_AfterOpen;
            Manager.OnStatusChanged += Manager_OnStatusChanged;
            Manager.OnDeviceFound += Manager_OnDeviceFound;
            Manager.OnDiscoveringCompleted += Manager_OnDiscoveringCompleted;
            Manager.OnDiscoveringStarted += Manager_OnDiscoveringStarted;

            InitResult = Manager.Open();
            if (InitResult != wclErrors.WCL_E_SUCCESS)
                WriteLn("Failed to open Blkuetooth Manager: 0x" + InitResult.ToString("X8"));
            else
            {
                if (Manager.Count == 0)
                {
                    InitResult = wclBluetoothErrors.WCL_E_BLUETOOTH_HARDWARE_NOT_AVAILABLE;
                    WriteLn("No Bluetooth hardware was found");
                }

                if (InitResult != wclErrors.WCL_E_SUCCESS)
                    Manager.Close();
            }

            InitEvent.Set();
            if (InitResult == wclErrors.WCL_E_SUCCESS)
            {
                wclMessageBroadcaster.Wait(TermintaeEvent);
                Manager.Close();
            }

            Manager = null;
        }

        private static void ShowRadioMenu()
        {
            Boolean Terminated = false;
            do
            {
                Console.Clear();
                WriteLn("RADIO MENU: " + Radio.ApiName);
                WriteLn("  1 - Get information");
                WriteLn("  2 - Discoverable ON");
                WriteLn("  3 - Discoverable OFF");
                WriteLn("  4 - Connectable ON");
                WriteLn("  5 - Connectable OFF");
                WriteLn("  6 - Turn ON");
                WriteLn("  7 - Turn OFF");
                WriteLn("  8 - Discover LE");
                WriteLn("  9 - Discover Classic");
                WriteLn("  0 - Enum connected");
                WriteLn("  e - Exit");
                WriteLn("Select option: ");
                String s = Console.ReadLine();
                if (s != "e")
                {
                    try
                    {
                        Byte Item = Convert.ToByte(s);
                        switch (Item)
                        {
                            case 0:
                                EnumConnected();
                                break;
                            case 1:
                                GetRadioInformation();
                                break;
                            case 2:
                                SetDiscoverable(true);
                                break;
                            case 3:
                                SetDiscoverable(false);
                                break;
                            case 4:
                                SetConnectable(true);
                                break;
                            case 5:
                                SetConnectable(false);
                                break;
                            case 6:
                                SwitchRadio(true);
                                break;
                            case 7:
                                SwitchRadio(false);
                                break;
                            case 8:
                                Discover(true);
                                break;
                            case 9:
                                Discover(false);
                                break;
                        }
                    }
                    catch
                    {
                    }
                }
                else
                    Terminated = true;
            } while (!Terminated);
        }

        static void Main(string[] args)
        {
            wclMessageBroadcaster.SetSyncMethod(wclMessageSynchronizationKind.skApc);

            try { CompleteEvent = new AutoResetEvent(false); } catch { CompleteEvent = null; }
            if (CompleteEvent == null)
                WriteLn("Create completion event failed");
            else
            {
                try { TermintaeEvent = new ManualResetEvent(false); } catch { TermintaeEvent = null; }
                if (TermintaeEvent == null)
                    WriteLn("Create termination event failed");
                else
                {
                    try { InitEvent = new ManualResetEvent(false); } catch { InitEvent = null; }
                    if (InitEvent == null)
                        WriteLn("Create initialization event failed");
                    else
                    {
                        InitResult = wclErrors.WCL_E_SUCCESS;
                        try { RunThread = new Thread(ManagerThread); } catch { RunThread = null; }
                        if (RunThread == null)
                            WriteLn("Create thread failed");
                        else
                        {
                            RunThread.Start();
                            InitEvent.WaitOne();
                            if (InitResult == wclErrors.WCL_E_SUCCESS)
                            {
                                Boolean Terminated = false;
                                do
                                {
                                    Console.Clear();
                                    WriteLn("BLUETOOTH MANAGER DEMO");
                                    WriteLn("======================");

                                    for (int i = 0; i < Manager.Count; i++)
                                        WriteLn((i + 1).ToString() + " - " + Manager[i].ApiName);
                                    WriteLn("0 - Exit");
                                    WriteLn("Select Radio");

                                    String s = Console.ReadLine();
                                    try
                                    {
                                        Byte Index = Convert.ToByte(s);
                                        if (Index == 0)
                                            Terminated = true;
                                        else
                                        {
                                            Index--;
                                            if (Index >= Manager.Count)
                                            {
                                                WriteLn("Wrong selection");
                                                WriteLn("Press any key to continue");
                                                Console.ReadKey();
                                            }
                                            else
                                            {
                                                Radio = Manager[Index];
                                                ShowRadioMenu();
                                            }
                                        }
                                    }
                                    catch
                                    {
                                        WriteLn("Wrong selection");
                                        WriteLn("Press any key to continue");
                                        Console.ReadKey();
                                    }
                                } while (!Terminated);

                                TermintaeEvent.Set();
                            }
                            RunThread.Join();
                        }

                        InitEvent.Close();
                        InitEvent = null;
                    }
                    TermintaeEvent.Close();
                    TermintaeEvent = null;
                }
                CompleteEvent.Close();
                CompleteEvent = null;
            }
        }
    }
}
