﻿using System;
using System.Threading;
using System.Collections.Generic;

using wclCommon;
using wclBluetooth;

namespace GattClient
{
    class Program
    {
        // The class incapsulates a single GATT connection.
        private class GattConnection
        {
            private ManualResetEvent ConnectEvent;
            private ManualResetEvent TerminateEvent;
            private ManualResetEvent InitEvent;
            private Int32 InitResult;
            private Thread GattThread;
            private wclGattClient Client;
            private wclBluetoothRadio BtRadio;

            private void GetMaxPduSize()
            {
                UInt16 Size;
                Int32 Res = Client.GetMaxPduSize(out Size);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: Get max PDU size error: 0x" + Res.ToString("X8"));
                else
                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: Max PDU size: " + Size.ToString());
            }

            private void GetConnectionParams()
            {
                wclBluetoothLeConnectionParameters Params;
                Int32 Res = Client.GetConnectionParams(out Params);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: Get connection params error: 0x" + Res.ToString("X8"));
                else
                {
                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: connection params");
                    Console.WriteLine("  Interval     : " + Params.Interval.ToString());
                    Console.WriteLine("  Latency      : " + Params.Latency.ToString());
                    Console.WriteLine("  Link timeout : " + Params.LinkTimeout.ToString());
                }
            }

            private void GetConnectionPhy()
            {
                wclBluetoothLeConnectionPhy Phy;
                Int32 Res = Client.GetConnectionPhyInfo(out Phy);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: Get connection PHY error: 0x" + Res.ToString("X8"));
                else
                {
                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: connection PHY");
                    Console.WriteLine("  Receive");
                    Console.WriteLine("    IsCoded        : " + Phy.Receive.IsCoded.ToString());
                    Console.WriteLine("    IsUncoded1MPhy : " + Phy.Receive.IsUncoded1MPhy.ToString());
                    Console.WriteLine("    IsUncoded2MPhy : " + Phy.Receive.IsUncoded2MPhy.ToString());
                    Console.WriteLine("  Transmit");
                    Console.WriteLine("    IsCoded        : " + Phy.Transmit.IsCoded.ToString());
                    Console.WriteLine("    IsUncoded1MPhy : " + Phy.Transmit.IsUncoded1MPhy.ToString());
                    Console.WriteLine("    IsUncoded2MPhy : " + Phy.Transmit.IsUncoded2MPhy.ToString());
                }
            }

            private void Client_OnCharacteristicChanged(Object Sender, UInt16 Handle, Byte[] Value)
            {
                Console.WriteLine("[" + Client.Address.ToString("X12") + "]: value " + Handle.ToString("X4") + " changed");
            }

            private void Client_OnDisconnect(Object Sender, Int32 Reason)
            {
                Console.WriteLine("[" + Client.Address.ToString("X12") + "]: disconnected with reason: 0x" + Reason.ToString("X8"));
            }

            private void Client_OnConnect(Object Sender, Int32 Error)
            {
                InitResult = Error;
                if (Error == wclErrors.WCL_E_SUCCESS)
                {
                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: connected");
                    GetMaxPduSize();
                    GetConnectionParams();
                    GetConnectionPhy();
                }
                else
                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: connect error: 0x" + Error.ToString("X8"));
                ConnectEvent.Set();
            }

            private void Client_OnMaxPduSizeChanged(object sender, EventArgs e)
            {
                GetMaxPduSize();
            }

            private void Client_OnConnectionParamsChanged(object sender, EventArgs e)
            {
                GetConnectionParams();
            }

            private void Client_OnConnectionPhyChanged(object sender, EventArgs e)
            {
                GetConnectionPhy();
            }

            private String UuidToString(wclGattUuid Uuid)
            {
                if (Uuid.IsShortUuid)
                    return Uuid.ShortUuid.ToString("X4");
                return Uuid.LongUuid.ToString();
            }

            private void DumpValue(Byte[] Value)
            {
                if (Value != null && Value.Length > 0)
                {
                    String s = "";
                    foreach (Byte b in Value)
                        s = s + b.ToString("X1");
                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: value: " + s);
                }
            }

            private void ReadCharacteristics(wclGattService Service)
            {
                wclGattCharacteristic[] Characteristics;
                Int32 Res = Client.ReadCharacteristics(Service, wclGattOperationFlag.goNone, out Characteristics);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: read characteristics error: 0x" + Res.ToString("X8"));
                else
                {
                    if (Characteristics == null || Characteristics.Length == 0)
                        Console.WriteLine("[" + Client.Address.ToString("X12") + "]: no characteristics found");
                    else
                    {
                        foreach (wclGattCharacteristic Characteristic in Characteristics)
                        {
                            Console.WriteLine("[" + Client.Address.ToString("X12") + "]: characteristic: " + UuidToString(Characteristic.Uuid));
                            if (Characteristic.IsReadable)
                            {
                                Console.WriteLine("[" + Client.Address.ToString("X12") + "]: readable");
                                Byte[] Value;
                                Res = Client.ReadCharacteristicValue(Characteristic, wclGattOperationFlag.goNone, out Value);
                                if (Value == null || Value.Length == 0)
                                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: value is empty");
                                else
                                    DumpValue(Value);
                            }

                            if (Characteristic.IsIndicatable || Characteristic.IsNotifiable)
                            {
                                Console.WriteLine("[" + Client.Address.ToString("X12") + "]: notifiable (or indicatable)");
                                wclGattCharacteristic Chr = Characteristic;
                                if (Characteristic.IsIndicatable && Characteristic.IsNotifiable)
                                    Chr.IsIndicatable = false;
                                Res = Client.Subscribe(Chr);
                                if (Res != wclErrors.WCL_E_SUCCESS)
                                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: subscribe error: 0x" + Res.ToString("X8"));
                                else
                                {
                                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: subscribed");
                                    Res = Client.WriteClientConfiguration(Chr, true, wclGattOperationFlag.goNone);
                                    if (Res != wclErrors.WCL_E_SUCCESS)
                                    {
                                        Console.WriteLine("[" + Client.Address.ToString("X12") + "]: write configuration error: 0x" + Res.ToString("X8"));
                                        Client.Unsubscribe(Chr);
                                    }
                                    else
                                        Console.WriteLine("[" + Client.Address.ToString("X12") + "]: write configuration completed");
                                }
                            }
                        }
                    }
                }
            }

            private void ReadServices()
            {
                wclGattService[] Services;
                Int32 Res = Client.ReadServices(wclGattOperationFlag.goNone, out Services);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Console.WriteLine("[" + Client.Address.ToString("X12") + "]: read services error: 0x" + Res.ToString("X8"));
                else
                {
                    if (Services == null || Services.Length == 0)
                        Console.WriteLine("[" + Client.Address.ToString("X12") + "]: no services found");
                    else
                    {
                        Console.WriteLine("[" + Client.Address.ToString("X12") + "]: found " + Services.Length + " services");
                        foreach (wclGattService Service in Services)
                        {
                            Console.WriteLine("[" + Client.Address.ToString("X12") + "]: service: " + UuidToString(Service.Uuid));
                            ReadCharacteristics(Service);
                        }
                    }
                }
            }

            private void ConnectionThread()
            {
                InitResult = Client.Connect(BtRadio);
                if (InitResult == wclErrors.WCL_E_SUCCESS)
                    wclMessageBroadcaster.Wait(ConnectEvent);

                if (InitResult == wclErrors.WCL_E_SUCCESS)
                    ReadServices();

                InitEvent.Set();

                if (InitResult == wclErrors.WCL_E_SUCCESS)
                {
                    wclMessageBroadcaster.Wait(TerminateEvent);
                    Client.Disconnect();
                }
            }

            public GattConnection()
            {
                ConnectEvent = null;
                TerminateEvent = null;
                InitEvent = null;
                InitResult = wclErrors.WCL_E_SUCCESS;
                GattThread = null;
                Client = null;
                BtRadio = null;
            }

            public Boolean Connect(wclBluetoothRadio Radio, Int64 Address)
            {
                if (GattThread != null)
                    return false;

                Boolean Result = false;

                try { ConnectEvent = new ManualResetEvent(false); } catch { ConnectEvent = null; }
                if (ConnectEvent == null)
                    Console.WriteLine("[" + Address.ToString("X12") + "]: Create connection event failed");
                else
                {
                    try { TerminateEvent = new ManualResetEvent(false); } catch { TerminateEvent = null; }
                    if (TerminateEvent == null)
                        Console.WriteLine("[" + Address.ToString("X12") + "]: Create termination event failed");
                    else
                    {
                        try { InitEvent = new ManualResetEvent(false);  } catch { InitEvent = null; }
                        if (InitEvent == null)
                            Console.WriteLine("[" + Address.ToString("X12") + "]: Create initialization event failed");
                        else
                        {
                            try { GattThread = new Thread(ConnectionThread); } catch { GattThread = null; }
                            if (GattThread == null)
                                Console.WriteLine("[" + Address.ToString("X12") + "]: Create communication thread failed");
                            else
                            {
                                BtRadio = Radio;

                                Client = new wclGattClient();
                                Client.Address = Address;
                                Client.OnConnect += Client_OnConnect;
                                Client.OnDisconnect += Client_OnDisconnect;
                                Client.OnCharacteristicChanged += Client_OnCharacteristicChanged;
                                Client.OnMaxPduSizeChanged += Client_OnMaxPduSizeChanged;
                                Client.OnConnectionParamsChanged += Client_OnConnectionParamsChanged;
                                Client.OnConnectionPhyChanged += Client_OnConnectionPhyChanged;

                                InitResult = wclErrors.WCL_E_SUCCESS;
                                GattThread.Start();

                                InitEvent.WaitOne();
                                Result = (InitResult == wclErrors.WCL_E_SUCCESS);
                                if (!Result)
                                {
                                    GattThread.Join();
                                    GattThread = null;
                                    Client = null;
                                    BtRadio = null;
                                }
                            }

                            InitEvent.Close();
                            InitEvent = null;
                        }

                        if (!Result)
                        {
                            TerminateEvent.Close();
                            TerminateEvent = null;
                        }
                    }

                    if (!Result)
                    {
                        ConnectEvent.Close();
                        ConnectEvent = null;
                    }
                }
                return Result;
            }

            public void Disconnect()
            {
                if (GattThread != null)
                {
                    TerminateEvent.Set();
                    GattThread.Join();
                    
                    ConnectEvent.Close();
                    ConnectEvent = null;

                    TerminateEvent.Close();
                    TerminateEvent = null;

                    InitResult = wclErrors.WCL_E_SUCCESS;
                    GattThread = null;
                    Client = null;
                    BtRadio = null;
                }
            }
        }
        private static List<GattConnection> Connections = null;
        private static List<Int64> Devices = null;
        private static ManualResetEvent DiscoveryEvent = null;

        private static void Manager_OnDiscoveringCompleted(Object Sender, wclBluetoothRadio Radio, Int32 Error)
        {
            Console.WriteLine("Discovering compleetd with result: 0x" + Error.ToString("X8"));
            DiscoveryEvent.Set();
        }

        private static void Manager_OnDeviceFound(Object Sender, wclBluetoothRadio Radio, Int64 Address)
        {
            Console.WriteLine("Device found: " + Address.ToString("X12"));
            Devices.Add(Address);
        }

        private static void Manager_OnDiscoveringStarted(Object Sender, wclBluetoothRadio Radio)
        {
            Console.WriteLine("Discovering started");
            Devices.Clear();
        }

        static void Main(string[] args)
        {
            // Set synchronization.
            wclMessageBroadcaster.SetSyncMethod(wclMessageSynchronizationKind.skApc);
            
            Console.Clear();
            Console.WriteLine("GATT CLIENT DEMO");
            Console.WriteLine("================");
            Console.WriteLine("This demo discovers all BLE devices and connects to each of them.");
            Console.WriteLine("After connection established the demo trys to read services and characteristics.");
            Console.WriteLine("It also subscribes for characteristic changes notification.");
            Console.WriteLine("");

            DiscoveryEvent = new ManualResetEvent(false);

            Connections = new List<GattConnection>();
            Devices = new List<Int64>();

            // Create and initialize Bluetooth Manager.
            wclBluetoothManager Manager = new wclBluetoothManager();
            Manager.OnDiscoveringStarted += Manager_OnDiscoveringStarted;
            Manager.OnDeviceFound += Manager_OnDeviceFound;
            Manager.OnDiscoveringCompleted += Manager_OnDiscoveringCompleted;

            // Try to open Bluetooth Manager.
            Int32 Res = Manager.Open();
            if (Res != wclErrors.WCL_E_SUCCESS)
                Console.WriteLine("Bluetooth Manager open failed. Error: 0x" + Res.ToString("X8"));
            else
            {
                if (Manager.Count == 0)
                    Console.WriteLine("No Bluetooth hardware was found");
                else
                {
                    wclBluetoothRadio Radio = null;
                    // Try to find available Bluetooth Radio.
                    for (int i = 0; i < Manager.Count; i++)
                    {
                        if (Manager[i].Available)
                        {
                            Radio = Manager[i];
                            break;
                        }
                    }

                    if (Radio == null)
                        Console.WriteLine("No Bluetooth Radio was found");
                    else
                    {
                        Console.WriteLine("All ready");
                        Console.WriteLine("Press any key to start discovering Bluetooth LE devices");
                        Console.WriteLine("Press ENTER to start the application.");
                        Console.ReadKey();

                        // Start discovering.
                        Res = Radio.Discover(10, wclBluetoothDiscoverKind.dkBle);
                        if (Res != wclErrors.WCL_E_SUCCESS)
                            Console.WriteLine("Failed to start discovering. Error: 0x" + Res.ToString("X8"));
                        else
                        {
                            // Wait for completion. We must use WaitEx to be able to process APC.
                            wclMessageBroadcaster.Wait(DiscoveryEvent);

                            if (Devices.Count == 0)
                                Console.WriteLine("No Bluetooth LE devices were found");
                            else
                            {
                                // Try to connect to each found device.
                                foreach (Int64 Address in Devices)
                                {
                                    Console.WriteLine("Connect to " + Address.ToString("X12"));
                                    GattConnection Connection = new GattConnection();
                                    if (!Connection.Connect(Radio, Address))
                                        Console.WriteLine("Connect to failed");
                                    else
                                        Connections.Add(Connection);
                                }

                                Console.ReadLine();
                                foreach (GattConnection Connection in Connections)
                                    Connection.Disconnect();
                            }
                        }
                    }
                }
                Manager.Close();
            }
            Manager = null;

            Devices.Clear();
            Devices = null;
            Connections.Clear();
            Connections = null;

            DiscoveryEvent.Close();
            DiscoveryEvent = null;
        }
    }
}
