﻿using System;
using System.Threading;

using wclCommon;
using wclBluetooth;

namespace GattServer
{
    class Program
    {
        private const UInt16 SHORT_DATA_LEN = 5;
        private const UInt16 LONG_DATA_LEN = 1024;

        private static UInt32 FCounter;
        private static Byte[] FLongData;
        private static Byte[] FShortData;

        private static void Server_OnStarted(Object sender, EventArgs e)
        {
            Console.WriteLine("Server has been started");
            FCounter = 0;
        }

        private static void Server_OnStopped(Object sender, EventArgs e)
        {
            Console.WriteLine("Server has been stopped");
        }

        private static void Server_OnWrite(Object Sender, wclGattServerClient Client, wclGattLocalCharacteristic Characteristic,
            wclGattLocalCharacteristicWriteRequest Request)
        {
            Console.WriteLine("Data received from " + Client.Address.ToString("X12"));
            Console.WriteLine("  Size: " + Request.Data.Length.ToString());
            Console.WriteLine("  Offset: " + Request.Offset.ToString());

            foreach (Byte b in Request.Data)
                Console.Write(b.ToString("X2"));
            Console.WriteLine();

            if (Request.WithResponse)
            {
                Console.WriteLine("  Write With Response. Sending response");
                Int32 Res = Request.Respond();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Console.WriteLine("  Respond failed: 0x" + Res.ToString("X8"));
            }
        }

        private static void Server_OnRead(Object Sender, wclGattServerClient Client, wclGattLocalCharacteristic Characteristic,
            wclGattLocalCharacteristicReadRequest Request)
        {
            Console.WriteLine("Read request from " + Client.Address.ToString("X12"));
            Console.WriteLine("  Offset: " + Request.Offset.ToString());
            Console.WriteLine("  Read Buffer Size: " + Request.Size.ToString());

            Int32 Res;
            Boolean IsShort = (Characteristic.Uuid.IsShortUuid && Characteristic.Uuid.ShortUuid == 0xFFF2);
            if (IsShort)
                Res = Request.Respond(FShortData);
            else
            {
                Byte[] Tmp = new Byte[LONG_DATA_LEN - Request.Offset];
                Array.Copy(FLongData, Request.Offset, Tmp, 0, LONG_DATA_LEN - Request.Offset);
                Res = Request.Respond(Tmp);
            }
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                Console.WriteLine("  Set data failed: 0x" + Res.ToString("X8"));
                Console.WriteLine("  respond with error");
                Res = Request.RespondWithError(wclBluetoothErrors.WCL_E_BLUETOOTH_LE_UNLIKELY);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Console.WriteLine("  Respond failed: 0x" + Res.ToString("X8"));
            }
        }

        private static void Server_OnSubscribed(Object Sender, wclGattServerClient Client, wclGattLocalCharacteristic Characteristic)
        {
            String Str = "Client subscribed: " + Client.Address.ToString("X12");

            UInt16 Size;
            Int32 Res = Client.GetMaxNotificationSize(out Size);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Str = Str + " Get max notification size failed: 0x" + Res.ToString("X8");
            else
                Str = Str + " Max notification size: " + Size.ToString();
            Console.WriteLine(Str);

            Byte[] a = BitConverter.GetBytes(FCounter);
            Res = Characteristic.Notify(Client.Address, a);
            FCounter++;
            if (Res != wclErrors.WCL_E_SUCCESS)
                Console.WriteLine("Notification failed: " + Res.ToString("X8"));
        }

        private static void Server_OnUnsubscribed(Object Sender, wclGattServerClient Client, wclGattLocalCharacteristic Characteristic)
        {
            Console.WriteLine("Client unsubscribed :" + Client.Address.ToString("X12"));
        }

        private static void ServerThread(Object Event)
        {
            ManualResetEvent TerminateEVent = (ManualResetEvent)Event;

            // Crerate and open Bluetooth Manager.
            wclBluetoothManager Manager = new wclBluetoothManager();
            Int32 Res = Manager.Open();
            if (Res != wclErrors.WCL_E_SUCCESS)
                Console.WriteLine("Open Bluetooth Manager failed: 0x" + Res.ToString("X8"));
            else
            {
                if (Manager.Count == 0)
                    Console.WriteLine("No Bluetooth hardware was found");
                else
                {
                    wclBluetoothRadio Radio = null;
                    for (Int32 i = 0; i < Manager.Count; i++)
                    {
                        if (Manager[i].Available)
                        {
                            Radio = Manager[i];
                            break;
                        }
                    }
                    if (Radio == null)
                        Console.WriteLine("No available Bluetooth radio found");
                    else
                    {
                        // Create serve rand set its event handlers.
                        wclGattServer Server = new wclGattServer();
                        Server.OnStarted += Server_OnStarted;
                        Server.OnStopped += Server_OnStopped;
                        Server.OnWrite += Server_OnWrite;
                        Server.OnRead += Server_OnRead;
                        Server.OnSubscribed += Server_OnSubscribed;
                        Server.OnUnsubscribed += Server_OnUnsubscribed;
                        Server.OnClientConnected += Server_OnClientConnected;
                        Server.OnClientDisconnected += Server_OnClientDisconnected;
                        Server.OnNotificationSizeChanged += Server_OnNotificationSizeChanged;
                        Server.OnMaxPduSizeChanged += Server_OnMaxPduSizeChanged;
                        Server.OnConnectionParamsChanged += Server_OnConnectionParamsChanged;
                        Server.OnConnectionPhyChanged += Server_OnConnectionPhyChanged;

                        Res = Server.Initialize(Radio);
                        if (Res != wclErrors.WCL_E_SUCCESS)
                            Console.WriteLine("Initialize server failed: 0x" + Res.ToString("X8"));
                        else
                        {
                            Console.WriteLine("Server initialized");

                            Console.WriteLine("Preparing services");
                            wclGattUuid Uuid = new wclGattUuid();
                            Uuid.IsShortUuid = true;
                            Uuid.ShortUuid = 0xFFF0;

                            Console.WriteLine("Add service");
                            wclGattLocalService Service;

                            Res = Server.AddService(Uuid, out Service);
                            if (Res != wclErrors.WCL_E_SUCCESS)
                                Console.WriteLine("Failed to add service: 0x" + Res.ToString("X8"));
                            else
                            {
                                Console.WriteLine("Service has been added");

                                Console.WriteLine("Adding characteristics");
                                Uuid.IsShortUuid = true;

                                wclGattLocalCharacteristicParameters Params = new wclGattLocalCharacteristicParameters();
                                wclGattLocalCharacteristic Char;

                                Params.Descriptors = null;
                                Params.PresentationFormats = null;

                                Console.WriteLine("Add empty characteristic");
                                Uuid.ShortUuid = 0xFFF1;
                                Params.Props = wclGattLocalCharacteristicProperty.cpNone;
                                Params.UserDescription = "Empty characteristic";
                                Res = Service.AddCharacteristic(Uuid, Params, out Char);
                                if (Res != wclErrors.WCL_E_SUCCESS)
                                    Console.WriteLine("Failed to add empty characteristic: 0x" + Res.ToString("X8"));
                                else
                                {
                                    Console.WriteLine("Add short readable characteristic");
                                    Uuid.ShortUuid = 0xFFF2;
                                    Params.Props = wclGattLocalCharacteristicProperty.cpReadable;
                                    Params.UserDescription = "Short readable characteristic";
                                    // Add fake presentation format
                                    wclGattCharacteristicFormat Format = new wclGattCharacteristicFormat()
                                    {
                                        AUnit = 0x0101,
                                        Description = 0x0202,
                                        Exponent = 0x01,
                                        Format = wclGattPresentationFormat.fmtUInt32,
                                        NameSpace = 0x01
                                    };
                                    Params.PresentationFormats = new wclGattCharacteristicFormat[1];
                                    Params.PresentationFormats[0] = Format;
                                    Res = Service.AddCharacteristic(Uuid, Params, out Char);
                                    if (Res != wclErrors.WCL_E_SUCCESS)
                                        Console.WriteLine("Failed to add short readable characteristic: 0x" + Res.ToString("X8"));
                                    else
                                    {
                                        Console.WriteLine("Add long readable characteristic");
                                        Uuid.ShortUuid = 0xFFF3;
                                        Params.Props = wclGattLocalCharacteristicProperty.cpReadable;
                                        Params.UserDescription = "Long readable characteristic";
                                    Res= Service.AddCharacteristic(Uuid, Params, out Char);
                                        if (Res != wclErrors.WCL_E_SUCCESS)
                                            Console.WriteLine("Failed to add long readable characteristic: 0x" + Res.ToString("X8"));
                                        else
                                        {
                                            Console.WriteLine("Add writable characteristic");
                                            Uuid.ShortUuid = 0xFFF4;
                                            Params.Props = wclGattLocalCharacteristicProperty.cpWritable |
                                                wclGattLocalCharacteristicProperty.cpWritableWithoutResponse;
                                            Params.UserDescription = "Writable characteristic";
                                            // Add second face presentation format
                                            Format.AUnit = 0x0303;
                                            Format.Description = 0x0404;
                                            Format.Exponent = 0x02;
                                            Format.Format = wclGattPresentationFormat.fmtUInt48;
                                            Format.NameSpace = 0x00;
                                            Array.Resize(ref Params.PresentationFormats, 2);
                                            Params.PresentationFormats[1] = Format;
                                            // ADDED : END
                                            Res = Service.AddCharacteristic(Uuid, Params, out Char);
                                            if (Res != wclErrors.WCL_E_SUCCESS)
                                                Console.WriteLine("Failed to add writable characteristic: 0x" + Res.ToString("X8"));
                                            else
                                            {
                                                Console.WriteLine("Add notifiable characteristic");
                                                Uuid.ShortUuid = 0xFFF5;
                                                Params.Props = wclGattLocalCharacteristicProperty.cpNotifiable;
                                                Params.UserDescription = "Notifiable characteristic";
                                                Params.PresentationFormats = null;
                                                // Add custom descriptor
                                                Byte[] Data = { 0x01, 0x02 };
                                                wclGattLocalDescriptor Descriptor = new wclGattLocalDescriptor()
                                                {
                                                    Data = Data,
                                                    ReadProtectionLevel = wclGattProtectionLevel.plNone,
                                                    WriteProtectionLevel = wclGattProtectionLevel.plNone,
                                                    Uuid = 0x2908
                                                };
                                                Params.Descriptors = new wclGattLocalDescriptor[1];
                                                Params.Descriptors[0] = Descriptor;
                                                Res = Service.AddCharacteristic(Uuid, Params, out Char);
                                                if (Res != wclErrors.WCL_E_SUCCESS)
                                                    Console.WriteLine("Failed to add notifiable characteristic: 0x" + Res.ToString("X8"));
                                                else
                                                {
                                                    Console.WriteLine("Add indicatable characteristic");
                                                    Uuid.ShortUuid = 0xFFF6;
                                                    Params.Props = wclGattLocalCharacteristicProperty.cpIndicatable;
                                                    Params.UserDescription = "Indicatable characteristic";
                                                    Res = Service.AddCharacteristic(Uuid, Params, out Char);
                                                    if (Res != wclErrors.WCL_E_SUCCESS)
                                                        Console.WriteLine("Failed to add indicatable characteristic: 0x" + Res.ToString("X8"));
                                                    else
                                                    {
                                                        Console.WriteLine("All charactertistics were added");

                                                        Res = Server.Start();
                                                        if (Res != wclErrors.WCL_E_SUCCESS)
                                                            Console.WriteLine("Start server failed: 0x" + Res.ToString("X8"));
                                                        else
                                                        {
                                                            wclMessageBroadcaster.Wait(TerminateEVent);

                                                            Res = Server.Stop();
                                                            if (Res != wclErrors.WCL_E_SUCCESS)
                                                                Console.WriteLine("Stop server failed: 0x" + Res.ToString("X8"));
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }

                            Console.WriteLine("Uninitializing server");
                            Res = Server.Uninitialize();
                            if (Res != wclErrors.WCL_E_SUCCESS)
                                Console.WriteLine("Uninitialize server failed: 0x" + Res.ToString("X8"));

                            Server = null;
                        }
                    }
                }
                Manager.Close();
            }
            Manager = null;
        }

        private static void Server_OnConnectionPhyChanged(object Sender, wclGattServerClient Client)
        {
            GetConPhy(Client);
        }

        private static void Server_OnConnectionParamsChanged(object Sender, wclGattServerClient Client)
        {
            GetConParams(Client);
        }

        private static void Server_OnMaxPduSizeChanged(object Sender, wclGattServerClient Client)
        {
            UInt16 Size;
            Int32 Res = Client.GetMaxPduSize(out Size);
            if (Res == wclErrors.WCL_E_SUCCESS)
                Console.WriteLine(Client.Address.ToString("X12") + " PDU size changed: " + Size.ToString());
            else
                Console.WriteLine(Client.Address.ToString("X12") + " get max PDU size failed: 0x" + Res.ToString("X8"));
        }

        private static void Server_OnNotificationSizeChanged(object Sender, wclGattServerClient Client)
        {
            UInt16 Size;
            Int32 Res = Client.GetMaxNotificationSize(out Size);
            if (Res == wclErrors.WCL_E_SUCCESS)
            {
                Console.WriteLine(Client.Address.ToString("X12") + " notification size changed: " +
                    Size.ToString());
            }
            else
            {
                Console.WriteLine(Client.Address.ToString("X12") + " get max notification size failed: 0x" +
                    Res.ToString("X8"));
            }
        }

        private static void Server_OnClientDisconnected(object Sender, wclGattServerClient Client)
        {
            Console.WriteLine("Client disconnected :" + Client.Address.ToString("X12"));
        }

        private static void Server_OnClientConnected(object Sender, wclGattServerClient Client)
        {
            Console.WriteLine("Client connected :" + Client.Address.ToString("X12"));

            UInt16 Size;
            Int32 Res = Client.GetMaxPduSize(out Size);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Console.WriteLine("Get max PDU size failed: 0x" + Res.ToString("X8"));
            else
                Console.WriteLine("Max PDU size: " + Size.ToString());

            GetConParams(Client);
            GetConPhy(Client);
        }

        private static void GetConParams(wclGattServerClient Client)
        {
            wclBluetoothLeConnectionParameters Params;
            Int32 Res = Client.GetConnectionParams(out Params);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Console.WriteLine("Get connection params failed: 0x" + Res.ToString("X8"));
            else
            {
                Console.WriteLine("Connection params [" + Client.Address.ToString("X12") + "]");
                Console.WriteLine("  Interval     : " + Params.Interval.ToString());
                Console.WriteLine("  Latency      : " + Params.Latency.ToString());
                Console.WriteLine("  Link Timeout : " + Params.LinkTimeout.ToString());
            }
        }

        private static void GetConPhy(wclGattServerClient Client)
        {
            wclBluetoothLeConnectionPhy Phy;
            Int32 Res = Client.GetConnectionPhyInfo(out Phy);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Console.WriteLine("Get connection PHY failed: 0x" + Res.ToString("X8"));
            else
            {
                Console.WriteLine("Connection PHY [" + Client.Address.ToString("X12") + "]");
                Console.WriteLine("  Receive");
                Console.WriteLine("    IsCoded        : " + Phy.Receive.IsCoded.ToString());
                Console.WriteLine("    IsUncoded1MPhy : " + Phy.Receive.IsUncoded1MPhy.ToString());
                Console.WriteLine("    IsUncoded2MPhy : " + Phy.Receive.IsUncoded2MPhy.ToString());
                Console.WriteLine("  Transmit");
                Console.WriteLine("    IsCoded        : " + Phy.Transmit.IsCoded.ToString());
                Console.WriteLine("    IsUncoded1MPhy : " + Phy.Transmit.IsUncoded1MPhy.ToString());
                Console.WriteLine("    IsUncoded2MPhy : " + Phy.Transmit.IsUncoded2MPhy.ToString());
            }
        }

        static void Main(string[] args)
        {
            wclMessageBroadcaster.SetSyncMethod(wclMessageSynchronizationKind.skApc);

            FShortData = new Byte[SHORT_DATA_LEN];
            for (UInt16 i = 0; i < SHORT_DATA_LEN; i++)
                FShortData[i] = (Byte)(i + 1);
            FLongData = new Byte[LONG_DATA_LEN];
            for (UInt16 i = 0; i < LONG_DATA_LEN; i++)
                FLongData[i] = (Byte)i;

            Console.Clear();
            Console.WriteLine("The demo shows how to use GATT Server in IoT applications.");
            Console.WriteLine("The demo uses APC syncrhonization.");

            ManualResetEvent TerminateEvent = new ManualResetEvent(false);

            Thread GattThread = new Thread(ServerThread);
            GattThread.Start(TerminateEvent);

            Console.WriteLine("Server start completion. Press ENTER to exit");
            Console.ReadLine();

            TerminateEvent.Set();
            GattThread.Join();

            TerminateEvent.Close();
        }
    }}