﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using wclCommon;
using wclBluetooth;

namespace GattServer
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private const UInt16 SHORT_DATA_LEN = 5;
        private const UInt16 LONG_DATA_LEN = 1024;

        private wclBluetoothManager FManager;
        private wclGattServer FServer;
        private Boolean FStarted;
        private UInt32 FCounter;

        private Byte[] FLongData;
        private Byte[] FShortData;

        private void FManager_AfterOpen(object sender, EventArgs e)
        {
            lbLog.Items.Add("Bluetooth Manager has been opened");
        }

        private void FManager_BeforeClose(object sender, EventArgs e)
        {
            lbLog.Items.Add("Bluetooth Manager is closing");
        }

        private void FServer_OnStarted(object sender, EventArgs e)
        {
            lbLog.Items.Add("Server has been started");
            FStarted = true;
            FCounter = 0;
        }

        private void FServer_OnStopped(object sender, EventArgs e)
        {
            lbLog.Items.Add("Server has been stopped");
            FStarted = false;
        }

        private void FServer_OnWrite(object Sender, wclGattServerClient Client, wclGattLocalCharacteristic Characteristic,
            wclGattLocalCharacteristicWriteRequest Request)
        {
            lbLog.Items.Add("Data received from " + Client.Address.ToString("X12"));
            lbLog.Items.Add("  Size: " + Request.Data.Length.ToString());
            lbLog.Items.Add("  Offset: " + Request.Offset.ToString());

            String s = "";
            foreach (Byte b in Request.Data)
                s = s + b.ToString("X2");
            lbLog.Items.Add(s);

            if (Request.WithResponse)
            {
                lbLog.Items.Add("  Write With Response. Sending response");
                Int32 Res = Request.Respond();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbLog.Items.Add("  Respond failed: 0x" + Res.ToString("X8"));
            }
        }

        private void FServer_OnRead(object Sender, wclGattServerClient Client, wclGattLocalCharacteristic Characteristic,
            wclGattLocalCharacteristicReadRequest Request)
        {
            lbLog.Items.Add("Read request from " + Client.Address.ToString("X12"));
            lbLog.Items.Add("  Offset: " + Request.Offset.ToString());
            lbLog.Items.Add("  Read Buffer Size: " + Request.Size.ToString());

            Int32 Res;
            Boolean IsShort = (Characteristic.Uuid.IsShortUuid &&
                Characteristic.Uuid.ShortUuid == 0xFFF2);
            if (IsShort)
                Res = Request.Respond(FShortData);
            else
            {
                Byte[] Tmp = new Byte[LONG_DATA_LEN - Request.Offset];
                Array.Copy(FLongData, Request.Offset, Tmp, 0, LONG_DATA_LEN - Request.Offset);
                Res = Request.Respond(Tmp);
            }
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                lbLog.Items.Add("  Set data failed: 0x" + Res.ToString("X8"));
                lbLog.Items.Add("  respond with error");
                Res = Request.RespondWithError(wclBluetoothErrors.WCL_E_BLUETOOTH_LE_UNLIKELY);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbLog.Items.Add("  Respond failed: 0x" + Res.ToString("X8"));
            }
        }

        private void FServer_OnSubscribed(object Sender, wclGattServerClient Client, wclGattLocalCharacteristic Characteristic)
        {
            String Str = "Client subscribed: " + Client.Address.ToString("X12");
            
            UInt16 Size;
            Int32 Res = Client.GetMaxNotificationSize(out Size);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Str = Str + " Get max notification size failed: 0x" + Res.ToString("X8");
            else
                Str = Str + " Max notification size: " + Size.ToString();
            lbLog.Items.Add(Str);
            
            Byte[] a = BitConverter.GetBytes(FCounter);
            Res = Characteristic.Notify(Client.Address, a);
            FCounter++;
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbLog.Items.Add("Notification failed: " + Res.ToString("X8"));
        }

        private void FServer_OnUnsubscribed(object Sender, wclGattServerClient Client, wclGattLocalCharacteristic Characteristic)
        {
            lbLog.Items.Add("Client unsubscribed :" + Client.Address.ToString("X12"));
        }

        private Boolean InitBluetooth(out wclBluetoothRadio Radio)
        {
            Radio = null;

            FManager = new wclBluetoothManager();
            FManager.AfterOpen += new EventHandler(FManager_AfterOpen);
            FManager.BeforeClose += new EventHandler(FManager_BeforeClose);

            Boolean Result = false;

            Int32 Res = FManager.Open();
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbLog.Items.Add("Bluetooth Manager open failed: 0x" + Res.ToString("X8"));
            else
            {
                lbLog.Items.Add("Checking for working Radio");
                Res = FManager.GetLeRadio(out Radio);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbLog.Items.Add("Get workign radio failed: 0x" + Res.ToString("X8"));
                else
                {
                    lbLog.Items.Add("Use " + Radio.ApiName + " radio");
                    Result = true;
                }

                if (!Result)
                    FManager.Close();
            }

            if (!Result)
                FManager = null;

            return Result;
        }

        private void UninitBluetooth()
        {
            lbLog.Items.Add("Close Bluetooth manager");
            if (FManager.Active)
                FManager.Close();
            FManager = null;
        }

        private Boolean InitServer(wclBluetoothRadio Radio)
        {
            FServer = new wclGattServer();
            FServer.OnStarted += new EventHandler(FServer_OnStarted);
            FServer.OnStopped += new EventHandler(FServer_OnStopped);
            FServer.OnWrite += new wclGattLocalCharacteristicWriteEvent(FServer_OnWrite);
            FServer.OnRead += new wclGattLocalCharacteristicReadEvent(FServer_OnRead);
            FServer.OnSubscribed += new wclGattLocalCharacteristicSubscribedClientsChangedEvent(FServer_OnSubscribed);
            FServer.OnUnsubscribed += new wclGattLocalCharacteristicSubscribedClientsChangedEvent(FServer_OnUnsubscribed);
            FServer.OnClientConnected += new wclGattServerClientEvent(FServer_OnClientConnected);
            FServer.OnClientDisconnected += new wclGattServerClientEvent(FServer_OnClientDisconnected);
            FServer.OnNotificationSizeChanged += new wclGattServerClientEvent(FServer_OnNotificationSizeChanged);
            FServer.OnMaxPduSizeChanged += new wclGattServerClientEvent(FServer_OnMaxPduSizeChanged);
            FServer.OnConnectionParamsChanged += new wclGattServerClientEvent(FServer_OnConnectionParamsChanged);
            FServer.OnConnectionPhyChanged += new wclGattServerClientEvent(FServer_OnConnectionPhyChanged);
            
            lbLog.Items.Add("Initialize server");
            Int32 Res = FServer.Initialize(Radio);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                lbLog.Items.Add("Initialize server failed: 0x" + Res.ToString("X8"));
                FServer = null;
                return false;
            }

            lbLog.Items.Add("Server initialized");
            return true;
        }

        void FServer_OnConnectionPhyChanged(object Sender, wclGattServerClient Client)
        {
            GetConPhy(Client);
        }

        void FServer_OnConnectionParamsChanged(object Sender, wclGattServerClient Client)
        {
            GetConParams(Client);
        }

        void FServer_OnMaxPduSizeChanged(object Sender, wclGattServerClient Client)
        {
            UInt16 Size;
            Int32 Res = Client.GetMaxPduSize(out Size);
            if (Res == wclErrors.WCL_E_SUCCESS)
                lbLog.Items.Add(Client.Address.ToString("X12") + " PDU size changed: " + Size.ToString());
            else
                lbLog.Items.Add(Client.Address.ToString("X12") + " get max PDU size failed: 0x" + Res.ToString("X8"));
        }

        void FServer_OnNotificationSizeChanged(object Sender, wclGattServerClient Client)
        {
            UInt16 Size;
            Int32 Res = Client.GetMaxNotificationSize(out Size);
            if (Res == wclErrors.WCL_E_SUCCESS)
            {
                lbLog.Items.Add(Client.Address.ToString("X12") + " notification size changed: " +
                    Size.ToString());
            }
            else
            {
                lbLog.Items.Add(Client.Address.ToString("X12") + " get max notification size failed: 0x" +
                    Res.ToString("X8"));
            }
        }

        void FServer_OnClientDisconnected(object Sender, wclGattServerClient Client)
        {
            lbLog.Items.Add("Client disconnected :" + Client.Address.ToString("X12"));
        }

        void FServer_OnClientConnected(object Sender, wclGattServerClient Client)
        {
            lbLog.Items.Add("Client connected :" + Client.Address.ToString("X12"));

            UInt16 Size;
            Int32 Res = Client.GetMaxPduSize(out Size);
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbLog.Items.Add("Get max PDU size failed: 0x" + Res.ToString("X8"));
            else
                lbLog.Items.Add("Max PDU size: " + Size.ToString());

            GetConParams(Client);
            GetConPhy(Client);
        }

        private void UninitServer()
        {
            lbLog.Items.Add("Uninitializing server");
            Int32 Res = FServer.Uninitialize();
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbLog.Items.Add("Uninitialize server failed: 0x" + Res.ToString("X8"));

            FServer = null;
        }

        private Boolean AddCharacteristics(wclGattLocalService Service)
        {
            lbLog.Items.Add("Adding characteristics services");
            wclGattUuid Uuid = new wclGattUuid();
            Uuid.IsShortUuid = true;

            wclGattLocalCharacteristicParameters Params = new wclGattLocalCharacteristicParameters();
            wclGattLocalCharacteristic Char;

            Params.Descriptors = null;
            Params.PresentationFormats = null;

            lbLog.Items.Add("Add empty characteristic");
            Uuid.ShortUuid = 0xFFF1;
            Params.Props = wclGattLocalCharacteristicProperty.cpNone;
            Params.UserDescription = "Empty characteristic";
            Int32 Res = Service.AddCharacteristic(Uuid, Params, out Char);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                lbLog.Items.Add("Failed to add empty characteristic: 0x" + Res.ToString("X8"));
                return false;
            }

            lbLog.Items.Add("Add short readable characteristic");
            Uuid.ShortUuid = 0xFFF2;
            Params.Props = wclGattLocalCharacteristicProperty.cpReadable;
            Params.UserDescription = "Short readable characteristic";
            Res = Service.AddCharacteristic(Uuid, Params, out Char);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                lbLog.Items.Add("Failed to add short readable characteristic: 0x" + Res.ToString("X8"));
                return false;
            }

            lbLog.Items.Add("Add long readable characteristic");
            Uuid.ShortUuid = 0xFFF3;
            Params.Props = wclGattLocalCharacteristicProperty.cpReadable;
            Params.UserDescription = "Long readable characteristic";
            Res = Service.AddCharacteristic(Uuid, Params, out Char);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                lbLog.Items.Add("Failed to add long readable characteristic: 0x" + Res.ToString("X8"));
                return false;
            }

            lbLog.Items.Add("Add writable characteristic");
            Uuid.ShortUuid = 0xFFF4;
            Params.Props = wclGattLocalCharacteristicProperty.cpWritable | wclGattLocalCharacteristicProperty.cpWritableWithoutResponse;
            Params.UserDescription = "Writable characteristic";
            Res = Service.AddCharacteristic(Uuid, Params, out Char);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                lbLog.Items.Add("Failed to add writable characteristic: 0x" + Res.ToString("X8"));
                return false;
            }

            lbLog.Items.Add("Add notifiable characteristic");
            Uuid.ShortUuid = 0xFFF5;
            Params.Props = wclGattLocalCharacteristicProperty.cpNotifiable;
            Params.UserDescription = "Notifiable characteristic";
            Res = Service.AddCharacteristic(Uuid, Params, out Char);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                lbLog.Items.Add("Failed to add notifiable characteristic: 0x" + Res.ToString("X8"));
                return false;
            }

            lbLog.Items.Add("Add indicatable characteristic");
            Uuid.ShortUuid = 0xFFF6;
            Params.Props = wclGattLocalCharacteristicProperty.cpIndicatable;
            Params.UserDescription = "Indicatable characteristic";
            Res = Service.AddCharacteristic(Uuid, Params, out Char);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                lbLog.Items.Add("Failed to add indicatable characteristic: 0x" + Res.ToString("X8"));
                return false;
            }

            lbLog.Items.Add("All charactertistics were added");
            return true;
        }

        private Boolean AddServices()
        {
            lbLog.Items.Add("Preparing services");
            wclGattUuid Uuid = new wclGattUuid();
            Uuid.IsShortUuid = true;
            Uuid.ShortUuid = 0xFFF0;

            lbLog.Items.Add("Add service");
            wclGattLocalService Service;
            Int32 Res = FServer.AddService(Uuid, out Service);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                lbLog.Items.Add("Failed to add service: 0x" + Res.ToString("X8"));
                return false;
            }

            lbLog.Items.Add("Service has been added");
            return AddCharacteristics(Service);
        }

        private void Start()
        {
            if (FStarted)
                lbLog.Items.Add("Already started");
            else
            {
                wclBluetoothRadio Radio = null;
                if (InitBluetooth(out Radio))
                {
                    Boolean Result = InitServer(Radio);
                    if (Result)
                    {
                        Result = AddServices();
                        if (Result)
                        {
                            lbLog.Items.Add("Starting server");
                            int Res = FServer.Start();
                            if (Res != wclErrors.WCL_E_SUCCESS)
                            {
                                lbLog.Items.Add("Start server failed: 0x" + Res.ToString("X8"));
                                Result = false;
                            }
                        }
                        if (!Result)
                            UninitServer();
                    }
                    if (!Result)
                        UninitBluetooth();
                }
            }
        }

        private void Stop()
        {
            if (!FStarted)
                lbLog.Items.Add("Not started");
            else
            {
                lbLog.Items.Add("Stopping server");
                Int32 Res = FServer.Stop();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbLog.Items.Add("Stop server failed: 0x" + Res.ToString("X8"));

                UninitServer();
                UninitBluetooth();
            }
        }

        public MainWindow()
        {
            InitializeComponent();
        }

        private void btStart_Click(object sender, RoutedEventArgs e)
        {
            Start();
        }

        private void btStop_Click(object sender, RoutedEventArgs e)
        {
            Stop();
        }

        private void fmMain_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            Stop();
        }

        private void fmMain_Loaded(object sender, RoutedEventArgs e)
        {
            FStarted = false;

            FShortData = new Byte[SHORT_DATA_LEN];
            for (UInt16 i = 0; i < SHORT_DATA_LEN; i++)
                FShortData[i] = (Byte)(i + 1);
            FLongData = new Byte[LONG_DATA_LEN];
            for (UInt16 i = 0; i < LONG_DATA_LEN; i++)
                FLongData[i] = (Byte)i;
        }

        private void GetConParams(wclGattServerClient Client)
        {
            wclBluetoothLeConnectionParameters Params;
            Int32 Res = Client.GetConnectionParams(out Params);
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbLog.Items.Add("Get connection params failed: 0x" + Res.ToString("X8"));
            else
            {
                lbLog.Items.Add("Connection params [" + Client.Address.ToString("X12") + "]");
                lbLog.Items.Add("  Interval     : " + Params.Interval.ToString());
                lbLog.Items.Add("  Latency      : " + Params.Latency.ToString());
                lbLog.Items.Add("  Link Timeout : " + Params.LinkTimeout.ToString());
            }
        }

        private void GetConPhy(wclGattServerClient Client)
        {
            wclBluetoothLeConnectionPhy Phy;
            Int32 Res = Client.GetConnectionPhyInfo(out Phy);
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbLog.Items.Add("Get connection PHY failed: 0x" + Res.ToString("X8"));
            else
            {
                lbLog.Items.Add("Connection PHY [" + Client.Address.ToString("X12") + "]");
                lbLog.Items.Add("  Receive");
                lbLog.Items.Add("    IsCoded        : " + Phy.Receive.IsCoded.ToString());
                lbLog.Items.Add("    IsUncoded1MPhy : " + Phy.Receive.IsUncoded1MPhy.ToString());
                lbLog.Items.Add("    IsUncoded2MPhy : " + Phy.Receive.IsUncoded2MPhy.ToString());
                lbLog.Items.Add("  Transmit");
                lbLog.Items.Add("    IsCoded        : " + Phy.Transmit.IsCoded.ToString());
                lbLog.Items.Add("    IsUncoded1MPhy : " + Phy.Transmit.IsUncoded1MPhy.ToString());
                lbLog.Items.Add("    IsUncoded2MPhy : " + Phy.Transmit.IsUncoded2MPhy.ToString());
            }
        }

        private void btClear_Click(object sender, RoutedEventArgs e)
        {
            lbLog.Items.Clear();
        }
    }
}
