﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using wclCommon;
using wclBluetooth;
using wclCommunication;

namespace RfCommClient
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private wclBluetoothManager Manager;
        private wclRfCommClient Client;

        public MainWindow()
        {
            InitializeComponent();
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            Manager = new wclBluetoothManager();
            Client = new wclRfCommClient();

            Manager.OnNumericComparison += new wclBluetoothNumericComparisonEvent(Manager_OnNumericComparison);
            Manager.OnPasskeyNotification += new wclBluetoothPasskeyNotificationEvent(Manager_OnPasskeyNotification);
            Manager.OnPasskeyRequest += new wclBluetoothPasskeyRequestEvent(Manager_OnPasskeyRequest);
            Manager.OnPinRequest += new wclBluetoothPinRequestEvent(Manager_OnPinRequest);
            Manager.OnConfirm += new wclBluetoothConfirmEvent(Manager_OnConfirm);
            Manager.OnAuthenticationCompleted += new wclBluetoothDeviceResultEvent(Manager_OnAuthenticationCompleted);
            Manager.OnDeviceFound += new wclBluetoothDeviceEvent(Manager_OnDeviceFound);
            Manager.OnDiscoveringCompleted += new wclBluetoothResultEvent(Manager_OnDiscoveringCompleted);
            Manager.OnDiscoveringStarted += new wclBluetoothEvent(Manager_OnDiscoveringStarted);

            Client.OnConnect += new wclClientConnectionConnectEvent(Client_OnConnect);
            Client.OnDisconnect += new wclClientConnectionDisconnectEvent(Client_OnDisconnect);
            Client.OnData += new wclConnectionDataEvent(Client_OnData);

            // In real application you should always analize the result code.
            // In this demo we assume that all is always OK.
            Manager.Open();

            cbAuthentication.IsChecked = Client.Authentication;
            cbEncryption.IsChecked = Client.Encryption;
        }

        void Manager_OnAuthenticationCompleted(object Sender, wclBluetoothRadio Radio, long Address, int Error)
        {
            lbEvents.Items.Add("Authentication completed: " + Error.ToString("X8"));
        }

        void Manager_OnConfirm(object Sender, wclBluetoothRadio Radio, long Address, out bool Confirm)
        {
            Confirm = true;
            lbEvents.Items.Add("Just works pairing");
        }

        void Manager_OnNumericComparison(object Sender, wclBluetoothRadio Radio, long Address, uint Number, out bool Confirm)
        {
            // Accept any pairing.
            Confirm = true;
            lbEvents.Items.Add("Numeric comparison: " + Number.ToString());
        }

        void Manager_OnPasskeyNotification(object Sender, wclBluetoothRadio Radio, long Address, uint Passkey)
        {
            lbEvents.Items.Add("Passkey notification: " + Passkey.ToString());
        }

        void Manager_OnPasskeyRequest(object Sender, wclBluetoothRadio Radio, long Address, out uint Passkey)
        {
            // Use 123456 as passkey.
            Passkey = 123456;
            lbEvents.Items.Add("Passkey request: " + Passkey.ToString());
        }

        void Manager_OnPinRequest(object Sender, wclBluetoothRadio Radio, long Address, out string Pin)
        {
            // Use '0000' as PIN.
            Pin = "0000";
            lbEvents.Items.Add("Pin request: " + Pin);
        }

        internal class DeviceItem
        {
            public String Address { get; set; }
            public String Name { get; set; }
            public String Type { get; set; }
            internal wclBluetoothRadio Radio { get; set; }
        };

        void Manager_OnDeviceFound(object Sender, wclBluetoothRadio Radio, long Address)
        {
            wclBluetoothDeviceType DevType = wclBluetoothDeviceType.dtMixed;
            Int32 Res = Radio.GetRemoteDeviceType(Address, out DevType);

            DeviceItem Item = new DeviceItem();
            Item.Address = Address.ToString("X12");
            Item.Name = "";
            Item.Radio = Radio; // To use it later.

            if (Res != wclErrors.WCL_E_SUCCESS)
                Item.Type = "Error: 0x" + Res.ToString("X8");
            else
            {
                switch (DevType)
                {
                    case wclBluetoothDeviceType.dtClassic:
                        Item.Type = "Classic";
                        break;
                    case wclBluetoothDeviceType.dtBle:
                        Item.Type = "BLE";
                        break;
                    case wclBluetoothDeviceType.dtMixed:
                        Item.Type = "Mixed";
                        break;
                    default:
                        Item.Type = "Unknown";
                        break;
                }
            }

            lvDevices.Items.Add(Item);

            lbEvents.Items.Add("Device found: " + Address.ToString("X12"));
        }

        void Manager_OnDiscoveringCompleted(object Sender, wclBluetoothRadio Radio, int Error)
        {
            if (lvDevices.Items.Count == 0)
                MessageBox.Show("No devices were found.", "Information", MessageBoxButton.OK, MessageBoxImage.Information);
            else
            {
                // Here we can update found devices names.
                foreach (DeviceItem Item in lvDevices.Items)
                {
                    Int64 Address = Convert.ToInt64(Item.Address, 16);
                    String DevName = "";
                    Int32 Res = Radio.GetRemoteName(Address, out DevName);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        Item.Name = "Error: 0x" + Res.ToString("X8");
                    else
                        Item.Name = DevName;
                }

                lvDevices.Items.Refresh();
            }

            lbEvents.Items.Add("Discovering completed");
        }

        void Manager_OnDiscoveringStarted(object Sender, wclBluetoothRadio Radio)
        {
            lvDevices.Items.Clear();
            lbEvents.Items.Add("Discovering started");
        }

        void Client_OnConnect(object Sender, int Error)
        {
            if (Error == wclErrors.WCL_E_SUCCESS)
            {
                lbEvents.Items.Add("Connected");

                GetBuffers();
            }
            else
                lbEvents.Items.Add("Connect error: 0x" + Error.ToString("X8"));
        }

        void Client_OnDisconnect(object Sender, int Reason)
        {
            lbEvents.Items.Add("Disconnected: 0x" + Reason.ToString("X8"));

            edReadBuffer.Text = "0";
            edWriteBuffer.Text = "0";
        }

        void Client_OnData(object Sender, byte[] Data)
        {
            if (Data != null && Data.Length > 0)
            {
                String Str = Encoding.ASCII.GetString(Data);

                lbEvents.Items.Add("Received " + Data.Length.ToString() + " bytes: " +
                    Str);
            }
        }

        private wclBluetoothRadio GetRadio()
        {
            wclBluetoothRadio Radio;
            Int32 Res = Manager.GetClassicRadio(out Radio);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"));
                return null;
            }
            return Radio;
        }

        private void Window_Closed(object sender, EventArgs e)
        {
            Client.Disconnect();
            Client = null;

            Manager.Close();
            Manager = null;
        }

        private void btDiscover_Click(object sender, RoutedEventArgs e)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                if (cbEnumPaired.IsChecked.Value)
                {
                    Int64[] Devices;
                    Int32 Res = Radio.EnumPairedDevices(out Devices);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        MessageBox.Show("Enum paired devices failed: 0x" + Res.ToString("X8"));
                    else
                    {
                        lvDevices.Items.Clear();
                        if (Devices != null && Devices.Length > 0)
                        {
                            foreach (Int64 Device in Devices)
                                Manager_OnDeviceFound(Manager, Radio, Device);
                            Manager_OnDiscoveringCompleted(Manager, Radio, wclErrors.WCL_E_SUCCESS);
                        }
                    }
                }
                else
                {
                    Int32 Res = Radio.Discover(10, wclBluetoothDiscoverKind.dkClassic);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        MessageBox.Show("Error starting discovering: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
        }

        private void btClear_Click(object sender, RoutedEventArgs e)
        {
            lbEvents.Items.Clear();
        }

        private void btConnect_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                wclBluetoothRadio Radio = GetRadio();
                if (Radio != null)
                {
                    Client.Address = Convert.ToInt64(((DeviceItem)lvDevices.SelectedItems[0]).Address, 16);
                    Client.Authentication = (Boolean)cbAuthentication.IsChecked;
                    Client.Encryption = (Boolean)cbEncryption.IsChecked;
                    Client.Timeout = (UInt32)Convert.ToInt32(edTimeout.Text);

                    Client.Channel = 0;
                    Client.Service = wclUUIDs.SerialPortServiceClass_UUID;
                    
                    Boolean Connect = false;
                    if ((Boolean)cbServiceName.IsChecked)
                    {
                        Guid Service = Client.Service;
                        wclBluetoothService[] Services;
                        Int32 Res = Radio.EnumRemoteServices(Client.Address, Service, out Services);
                        if (Res != wclErrors.WCL_E_SUCCESS)
                            MessageBox.Show("Service enumerating error: 0x" + Res.ToString("X8"));
                        else
                        {
                            if (Services == null || Services.Length == 0)
                                MessageBox.Show("Services not found");
                            else
                            {
                                foreach (wclBluetoothService s in Services)
                                {
                                    if (s.Name == edServiceName.Text)
                                    {
                                        Client.Channel = s.Channel;
                                        Connect = true;
                                        break;
                                    }
                                }
                                
                                if (!Connect)
                                    MessageBox.Show("Service not found");
                            }
                        }
                    }
                    else
                        Connect = true;

                    if (Connect)
                    {
                        Int32 Res = Client.Connect(Radio);
                        if (Res != wclErrors.WCL_E_SUCCESS)
                            MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                    }
                }
            }
        }

        private void btDisconnect_Click(object sender, RoutedEventArgs e)
        {
            Int32 Res = Client.Disconnect();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
        }

        private void btSend_Click(object sender, RoutedEventArgs e)
        {
            Byte[] Ansi = Encoding.ASCII.GetBytes(edText.Text);
            UInt32 Sent = 0;
            Int32 Res = Client.Write(Ansi, out Sent);
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"));
            else
                MessageBox.Show("Sent: " + Sent.ToString() + " from " + Ansi.Length.ToString());
        }

        private void btSetBuffers_Click(object sender, RoutedEventArgs e)
        {
            // Try to set read buffer size first.
            UInt32 Size = Convert.ToUInt32(edReadBuffer.Text);
            Int32 Res = Client.SetReadBufferSize(Size);
            if (Res == wclErrors.WCL_E_SUCCESS)
                lbEvents.Items.Add("Read buffer size changed.");
            else
                lbEvents.Items.Add("Set read buffer size error: 0x" + Res.ToString("X8"));

            // Now write buffer.
            Size = Convert.ToUInt32(edWriteBuffer.Text);
            Res = Client.SetWriteBufferSize(Size);
            if (Res == wclErrors.WCL_E_SUCCESS)
                lbEvents.Items.Add("Write buffer size changed.");
            else
                lbEvents.Items.Add("Set write buffer size error: 0x" + Res.ToString("X8"));
        }

        private void btGetBuffers_Click(object sender, RoutedEventArgs e)
        {
            GetBuffers();
        }

        private void GetBuffers()
        {
            // Try to get read and write buffers size.
            UInt32 Size = 0;
            Int32 Res = Client.GetReadBufferSize(out Size);
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbEvents.Items.Add("Get read buffer error: 0x" + Res.ToString("X8"));
            else
                edReadBuffer.Text = Size.ToString();

            Size = 0;
            Res = Client.GetWriteBufferSize(out Size);
            if (Res != wclErrors.WCL_E_SUCCESS)
                lbEvents.Items.Add("Get write buffer error: 0x" + Res.ToString("X8"));
            else
                edWriteBuffer.Text = Size.ToString();
        }

        private void btPair_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                wclBluetoothRadio Radio = GetRadio();
                if (Radio != null)
                {
                    Int32 Res = Radio.RemotePair(Convert.ToInt64(((DeviceItem)lvDevices.SelectedItems[0]).Address, 16),
                        wclBluetoothPairingMethod.pmClassic);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        MessageBox.Show("Pair failed: " + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Warning);
                }
            }
        }

        private void btUnpair_Click(object sender, RoutedEventArgs e)
        {
            if (lvDevices.SelectedItems.Count == 0)
                MessageBox.Show("Select device", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                wclBluetoothRadio Radio = GetRadio();
                if (Radio != null)
                {
                    Int32 Res = Radio.RemoteUnpair(Convert.ToInt64(((DeviceItem)lvDevices.SelectedItems[0]).Address, 16),
                        wclBluetoothPairingMethod.pmClassic);
                    if (Res != wclErrors.WCL_E_SUCCESS)
                        MessageBox.Show("Unpaired failed: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Warning);
                }
            }
        }
    }
}
