﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using wclCommon;
using wclBluetooth;
using wclCommunication;

namespace RfCommServer
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        private wclRfCommServer Server;
        private wclBluetoothManager Manager;

        public MainWindow()
        {
            InitializeComponent();
        }

        private void btClear_Click(object sender, RoutedEventArgs e)
        {
            lbEvents.Items.Clear();
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            Manager = new wclBluetoothManager();
            Server = new wclRfCommServer();

            Manager.OnNumericComparison += new wclBluetoothNumericComparisonEvent(Manager_OnNumericComparison);
            Manager.OnPasskeyNotification += new wclBluetoothPasskeyNotificationEvent(Manager_OnPasskeyNotification);
            Manager.OnPasskeyRequest += new wclBluetoothPasskeyRequestEvent(Manager_OnPasskeyRequest);
            Manager.OnPinRequest += new wclBluetoothPinRequestEvent(Manager_OnPinRequest);

            Server.OnClosed += new wclClientConnectionDisconnectEvent(Server_OnClosed);
            Server.OnConnect += new wclRfCommServerConnectEvent(Server_OnConnect);
            Server.OnData += new wclRfCommServerDataEvent(Server_OnData);
            Server.OnDisconnect += new wclRfCommServerDisconnectEvent(Server_OnDisconnect);
            Server.OnListen += new EventHandler(Server_OnListen);

            // In real application you should always analize the result code.
            // In this demo we assume that all is always OK.
            Manager.Open();

            cbAuthentication.IsChecked = Server.Authentication;
            cbEncryption.IsChecked = Server.Encryption;
        }

        void Manager_OnNumericComparison(object Sender, wclBluetoothRadio Radio, long Address, uint Number, out bool Confirm)
        {
            // Accept any pairing.
            Confirm = true;
            lbEvents.Items.Add("Numeric comparison: " + Number.ToString());
        }

        void Manager_OnPasskeyNotification(object Sender, wclBluetoothRadio Radio, long Address, uint Passkey)
        {
            lbEvents.Items.Add("Passkey notification: " + Passkey.ToString());
        }

        void Manager_OnPasskeyRequest(object Sender, wclBluetoothRadio Radio, long Address, out uint Passkey)
        {
            // Use 123456 as passkey.
            Passkey = 123456;
            lbEvents.Items.Add("Passkey request: " + Passkey.ToString());
        }

        void Manager_OnPinRequest(object Sender, wclBluetoothRadio Radio, long Address, out string Pin)
        {
            // Use '0000' as PIN.
            Pin = "0000";
            lbEvents.Items.Add("Pin request: " + Pin);
        }

        void Server_OnClosed(object Sender, int Reason)
        {
            lbEvents.Items.Add("Server closed on radio " +
                ((wclRfCommServer)Sender).Radio.ApiName + "with reason: 0x" + Reason.ToString("X8"));

            lvClients.Items.Clear();

            edReadBuffer.Text = "";
            edWriteBuffer.Text = "";
        }

        internal class ClientItem
        {
            public String Address { get; set; }
            public String Name { get; set; }

            internal wclRfCommServerClientConnection Client { get; set; }
        };

        void Server_OnConnect(object Sender, wclRfCommServerClientConnection client, int Error)
        {
            wclRfCommServerConnection Server = (wclRfCommServerConnection)client.Server;

            lbEvents.Items.Add("Client connected: " + client.Address.ToString("X12") +
                " on Radio " + Server.Radio.ApiName + " error: 0x" + Error.ToString("X8"));

            lvClients.Items.Add(new ClientItem {
                Address = client.Address.ToString("X12"),
                Name = GetDeviceName(client.Address),
                Client = client
            });
        }

        void Server_OnData(object Sender, wclRfCommServerClientConnection Client, byte[] Data)
        {
            if (Data != null && Data.Length > 0)
            {
                String Str = Encoding.ASCII.GetString(Data);

                wclRfCommServerConnection Server = (wclRfCommServerConnection)Client.Server;
                lbEvents.Items.Add("Received " + Data.Length.ToString() + " bytes: " + Str + " from " +
                    Client.Address.ToString("X12") + " on radio " + Server.Radio.ApiName);
            }
        }

        void Server_OnDisconnect(object Sender, wclRfCommServerClientConnection Client, int Reason)
        {
            wclRfCommServerConnection Server = (wclRfCommServerConnection)Client.Server;

            lbEvents.Items.Add("Client disconnected: " + Client.Address.ToString("X12") +
                " on radio " + Server.Radio.ApiName + " reason: 0x" + Reason.ToString("X8"));

            foreach (ClientItem Item in lvClients.Items)
            {
                if (Item.Client == Client)
                {
                    lvClients.Items.Remove(Item);
                    break;
                }
            }
        }

        void Server_OnListen(object sender, EventArgs e)
        {
            lbEvents.Items.Add("Server listening on radio " +
                ((wclRfCommServer)sender).Radio.ApiName + " on channel: " +
                ((wclRfCommServer)sender).AssignedChannel.ToString());
        }

        private void Window_Closed(object sender, EventArgs e)
        {
            Server.Close();
            Server = null;

            Manager.Close();
            Manager = null;
        }

        private wclBluetoothRadio GetRadio()
        {
            wclBluetoothRadio Radio;
            Int32 Res = Manager.GetClassicRadio(out Radio);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"));
                return null;
            }
            return Radio;
        }

        private void btListen_Click(object sender, RoutedEventArgs e)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                // Switch radio in discoverable and connectable mode.
                // Ignore any errors. In real application you need to check the result.
                Radio.SetDiscoverable(true);
                Radio.SetConnectable(true);

                Server.Authentication = (Boolean)cbAuthentication.IsChecked;
                Server.Encryption = (Boolean)cbEncryption.IsChecked;

                Server.Service = wclUUIDs.SerialPortServiceClass_UUID;
                Server.ServiceName = edServiceName.Text;

                Int32 Res = Server.Listen(Radio);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void btclose_Click(object sender, RoutedEventArgs e)
        {
            Int32 Res = Server.Close();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
        }

        private void btDisconnect_Click(object sender, RoutedEventArgs e)
        {
            if (lvClients.SelectedItems.Count == 0)
                MessageBox.Show("Select client", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                Int32 Res = ((ClientItem)lvClients.SelectedItems[0]).Client.Disconnect();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        private void btSend_Click(object sender, RoutedEventArgs e)
        {
            if (lvClients.SelectedItems.Count == 0)
                MessageBox.Show("Select client", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                Byte[] Ansi = Encoding.ASCII.GetBytes(edText.Text);
                UInt32 Sent = 0;
                Int32 Res = ((ClientItem)lvClients.SelectedItems[0]).Client.Write(Ansi, out Sent);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Error: 0x" + Res.ToString("X8"));
                else
                    MessageBox.Show("Sent: " + Sent.ToString() + " from " + Ansi.Length.ToString());
            }
        }

        private void btGetBuffers_Click(object sender, RoutedEventArgs e)
        {
            if (lvClients.SelectedItems.Count == 0)
                MessageBox.Show("Select client", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                wclRfCommServerClientConnection Client = ((ClientItem)lvClients.SelectedItems[0]).Client;

                // Get read buffer.
                UInt32 Size = 0;
                Int32 Res = Client.GetReadBufferSize(out Size);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbEvents.Items.Add("Get read buffer error: 0x" + Res.ToString("X8"));
                else
                    edReadBuffer.Text = Size.ToString();

                // Get write buffer.
                Size = 0;
                Res = Client.GetWriteBufferSize(out Size);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbEvents.Items.Add("Get write buffer error: 0x" + Res.ToString("X8"));
                else
                    edWriteBuffer.Text = Size.ToString();
            }
        }

        private void btSetBuffers_Click(object sender, RoutedEventArgs e)
        {
            if (lvClients.SelectedItems.Count == 0)
                MessageBox.Show("Select client", "Warning", MessageBoxButton.OK, MessageBoxImage.Warning);
            else
            {
                wclRfCommServerClientConnection Client = ((ClientItem)lvClients.SelectedItems[0]).Client;

                // Set read buffer.
                UInt32 Size = Convert.ToUInt32(edReadBuffer.Text);
                Int32 Res = Client.SetReadBufferSize(Size);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbEvents.Items.Add("Set read buffer error: 0x" + Res.ToString("X8"));

                // Set write buffer.
                Size = Convert.ToUInt32(edWriteBuffer.Text);
                Res = Client.SetWriteBufferSize(Size);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbEvents.Items.Add("Set write buffer error: 0x" + Res.ToString("X8"));
            }
        }

        private String GetDeviceName(Int64 Address)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                String DevName;
                Int32 Res = Radio.GetRemoteName(Address, out DevName);
                if (Res == wclErrors.WCL_E_SUCCESS)
                    return DevName;

                return "Error: 0x" + Res.ToString("X8");
            }

            return "";
        }
    }
}
