﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

using wclCommon;
using wclAudio;

namespace AudioVolume
{
    public partial class fmMain : Form
    {
        public class VerticalProgressBar : System.Windows.Forms.ProgressBar
        {
            protected override System.Windows.Forms.CreateParams CreateParams
            {
                get
                {
                    System.Windows.Forms.CreateParams cp = base.CreateParams;
                    cp.Style |= 0x04;
                    return cp;
                }
            }
        }
        
        private class AudioDeviceStorage
        {
            public wclAudioDevice Device;

            public override String ToString()
            {
                return Device.FriendlyName;
            }
        };

        private wclAudioSwitcher AudioSwitcher;
        private wclAudioMeter AudioMeter;
        private wclAudioVolume AudioVolume;
        private VerticalProgressBar[] FBars;
        
        public fmMain()
        {
            InitializeComponent();
        }

        private void fmMain_Load(Object sender, EventArgs e)
        {
            AudioSwitcher = new wclAudioSwitcher();
            AudioSwitcher.OnClosed += new EventHandler(AudioSwitcherClosed);
            AudioSwitcher.OnOpened += new EventHandler(AudioSwitcherOpened);
            AudioSwitcher.OnDeviceRemoved += new wclAudioDeviceEvent(AudioSwitcherDeviceRemoved);
            AudioSwitcher.OnDeviceAdded += new wclAudioDeviceEvent(AudioSwitcherDeviceAdded);
            AudioSwitcher.OnStateChanged += new wclAudioDeviceStateChangedEvent(AudioSwitcherStateChanged);
            AudioSwitcher.OnDefaultDeviceChanged += new wclAudioDefaultDeviceChangedEvent(AudioSwitcherDefaultDeviceChanged);

            AudioMeter = new wclAudioMeter();
            AudioMeter.OnOpened += new EventHandler(AudioMeterOpened);
            AudioMeter.OnClosed += new EventHandler(AudioMeterClosed);
            AudioMeter.OnDisconnected += new EventHandler(AudioMeterDisconnected);

            AudioVolume = new wclAudioVolume();
            AudioVolume.OnOpened += new EventHandler(AudioVolumeOpened);
            AudioVolume.OnClosed += new EventHandler(AudioVolumeClosed);
            AudioVolume.OnDisconnected += new EventHandler(AudioVolumeDisconnected);
            AudioVolume.OnChanged += new wclAudioVolumeChangedEvent(AudioVolumeChanged);

            FBars = null;
            
            Int32 Res = AudioSwitcher.Open();
            if (Res != wclErrors.WCL_E_SUCCESS)
                Trace("Audio switcher open failed", Res);
        }

        void AudioVolumeChanged(Object Sender, Boolean Muted, Single Volume, Single[] Volumes)
        {
            if (AudioVolume.Active)
            {
                Trace("Volume changed");

                cbMute.Checked = Muted;
                tbMaster.Value = (int)(Volume * 100);
            }
        }

        void AudioVolumeDisconnected(Object sender, EventArgs e)
        {
            Trace("Audio volume device " + AudioVolume.Id + " disconnected");
        }

        void AudioVolumeClosed(Object sender, EventArgs e)
        {
            Trace("Audio volume closed");
            cbEndpoints.Enabled = true;
            tbMaster.Visible = false;
            tbMaster.Value = 0;
            cbMute.Visible = false;
        }

        void AudioVolumeOpened(Object sender, EventArgs e)
        {
            Trace("Audio volume opened");
            cbEndpoints.Enabled = false;
            
            wclAudioHardwareFeature Features;
            Int32 Res = AudioVolume.GetHardwareFeatures(out Features);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Trace("Get hardware features failed", Res);
            else
            {
                Trace("Hardware features");
                if ((Features & wclAudioHardwareFeature.hfVolume) != 0)
                    Trace("  Volume");
                if ((Features & wclAudioHardwareFeature.hfMute) != 0)
                    Trace("  Mute");
                if ((Features & wclAudioHardwareFeature.hfMeter) != 0)
                    Trace("  Meter");
            }
            
            Single Volume;
            Res = AudioVolume.GetVolume(out Volume);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Trace("Get volume failed", Res);
            else
            {
                tbMaster.Visible = true;
                tbMaster.Value = (int)(Volume * 100);
            }
            
            Boolean Mute;
            Res = AudioVolume.GetMute(out Mute);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Trace("Get mute failed", Res);
            else
            {
                cbMute.Visible = true;
                cbMute.Checked = Mute;
            }
        }

        void AudioSwitcherDefaultDeviceChanged(Object Sender, String Id, wclAudioDeviceDataFlow Flow, wclAudioDeviceRole Role)
        {
            Trace("Active device changed");
            
            if (!AudioMeter.Active)
                SelectActiveDevice();
        }

        void AudioSwitcherStateChanged(Object Sender, String Id, wclAudioDeviceState State)
        {
            if (State == wclAudioDeviceState.asActive)
                AudioSwitcherDeviceAdded(AudioSwitcher, Id);
            else
                DeleteDevice(Id);
        }

        void AudioMeterDisconnected(Object sender, EventArgs e)
        {
            Trace("Audio meter device " + AudioMeter.Id + " disconnected");
        }

        void AudioMeterClosed(Object sender, EventArgs e)
        {
            Trace("Audio meter closed");
            Timer.Enabled = false;
            cbEndpoints.Enabled = true;

            if (FBars != null && FBars.Length > 0)
            {
                for (Int32 i = 0; i < FBars.Length; i++)
                {
                    Controls.Remove(FBars[i]);
                    FBars[i].Dispose();
                }
                FBars = null;
            }

            pbMaster.Value = 0;
            pbMaster.Visible = false;
        }

        void AudioMeterOpened(Object sender, EventArgs e)
        {
            Trace("Audio meter opened");
            cbEndpoints.Enabled = false;
            pbMaster.Visible = true;
            
            wclAudioHardwareFeature Features;
            Int32 Res = AudioMeter.GetHardwareFeatures(out Features);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Trace("Get hardware features failed", Res);
            else
            {
                Trace("Hardware features");
                if ((Features & wclAudioHardwareFeature.hfVolume) != 0)
                    Trace("  Volume");
                if ((Features & wclAudioHardwareFeature.hfMute) != 0)
                    Trace("  Mute");
                if ((Features & wclAudioHardwareFeature.hfMeter) != 0)
                    Trace("  Meter");
            }
            
            UInt32 Channels;
            Res = AudioMeter.GetChannels(out Channels);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Trace("Get channels count failed", Res);
            else
            {
                if (Channels == 0)
                {
                    Trace("No mettering channels available");
                    AudioMeter.Close();
                }
                else
                {
                    Trace("Channels: " + Channels.ToString());

                    Int32 Pos = pbMaster.Left + pbMaster.Width + 5;
                    FBars = new VerticalProgressBar[Channels];
                    for (UInt32 i = 0; i < Channels; i++)
                    {
                        VerticalProgressBar Bar = new VerticalProgressBar();
                        Bar.Width = pbMaster.Width;
                        Bar.Height = pbMaster.Height;
                        Bar.Top = pbMaster.Top;
                        Bar.Left = Pos;
                        Bar.Style = System.Windows.Forms.ProgressBarStyle.Continuous;
                        Pos = Bar.Left + Bar.Width + 5;
                        Controls.Add(Bar);

                        FBars[i] = Bar;
                    }
                    
                    UpdateChannelsValue();

                    Timer.Enabled = true;
                }
            }
        }

        void AudioSwitcherDeviceAdded(Object Sender, String Id)
        {
            wclAudioDevice Device;
            if (AudioSwitcher.GetDeviceDetails(Id, out Device) == wclErrors.WCL_E_SUCCESS)
                AddDevice(Device);
        }

        void AudioSwitcherDeviceRemoved(Object Sender, String Id)
        {
            DeleteDevice(Id);
        }

        void AudioSwitcherOpened(Object sender, EventArgs e)
        {
            Trace("Audio switcher opened");
            
            wclAudioDevice[] Devices;
            Int32 Res = AudioSwitcher.Enum(wclAudioDeviceState.asActive, out Devices);
            if (Res != wclErrors.WCL_E_SUCCESS)
                Trace("Enum audio endpoints failed", Res);
            else
            {
                if (Devices == null || Devices.Length == 0)
                    Trace("No active audio endpoints found");
                else
                {
                    foreach (wclAudioDevice Device in Devices)
                        AddDevice(Device);

                    if (cbEndpoints.Items.Count == 0)
                        Trace("No output audio endpoints found");
                    else
                        SelectActiveDevice();
                }
            }
        }

        void AudioSwitcherClosed(Object sender, EventArgs e)
        {
            Trace("Audio switcher closed");

            cbEndpoints.Items.Clear();
        }

        private void btClear_Click(Object sender, EventArgs e)
        {
            lbLog.Items.Clear();
        }

        private void Trace(String Msg)
        {
            lbLog.Items.Add(Msg);
        }

        private void Trace(String Msg, Int32 Error)
        {
            Trace(Msg + ": 0x" + Error.ToString("X8"));
        }

        private void DeleteDevice(String Id)
        {
            if (cbEndpoints.Items.Count > 0)
            {
                for (Int32 i = 0; i < cbEndpoints.Items.Count; i++)
                {
                    AudioDeviceStorage Storage = (AudioDeviceStorage)cbEndpoints.Items[i];
                    if (Storage.Device.Id == Id)
                    {
                        cbEndpoints.Items.RemoveAt(i);
                        if (cbEndpoints.Items.Count > 0)
                            SelectActiveDevice();
                        break;
                    }
                }
            }
        }

        private void SelectActiveDevice()
        {
            if (cbEndpoints.Items.Count > 0)
            {
                for (Int32 i = 0; i < cbEndpoints.Items.Count; i++)
                {
                    AudioDeviceStorage Storage = (AudioDeviceStorage)cbEndpoints.Items[i];
                    wclAudioDevice Device;
                    Int32 Res = AudioSwitcher.GetDeviceDetails(Storage.Device.Id, out Device);
                    if (Res == wclErrors.WCL_E_SUCCESS)
                        Storage.Device = Device;
                    if ((Storage.Device.Roles & wclAudioDeviceRole.drMultimedia) != 0)
                    {
                        cbEndpoints.SelectedIndex = i;
                        return;
                    }
                }
            }

            cbEndpoints.SelectedIndex = -1;
        }

        private void AddDevice(wclAudioDevice Device)
        {
            if (Device.State == wclAudioDeviceState.asActive && Device.Flow == wclAudioDeviceDataFlow.dfRender)
            {
                AudioDeviceStorage Storage = new AudioDeviceStorage();
                Storage.Device = Device;
                cbEndpoints.Items.Add(Storage);

                SelectActiveDevice();
            }
        }

        private void fmMain_FormClosed(Object sender, FormClosedEventArgs e)
        {
            AudioVolume.Close();
            AudioMeter.Close();
            AudioSwitcher.Close();
        }

        private void btClose_Click(Object sender, EventArgs e)
        {
            Int32 Res = AudioVolume.Close();
            if (Res != wclErrors.WCL_E_SUCCESS)
                Trace("Audio volume close failed", Res);

            Res = AudioMeter.Close();
            if (Res != wclErrors.WCL_E_SUCCESS)
                Trace("Audio meter close failed", Res);
        }

        private void btOpen_Click(object sender, EventArgs e)
        {
            if (cbEndpoints.SelectedIndex == -1)
                Trace("No audio endpoints found");
            else
            {
                String Id = ((AudioDeviceStorage)cbEndpoints.Items[cbEndpoints.SelectedIndex]).Device.Id;
                
                Int32 Res = AudioMeter.Open(Id);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Trace("Audio meter open failed", Res);

                Res = AudioVolume.Open(Id);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Trace("Audio volume open failed", Res);
            }
        }

        private void UpdateChannelsValue()
        {
            Single Value;
            Int32 Res = AudioMeter.GetPeak(out Value);
            if (Res == wclErrors.WCL_E_SUCCESS)
            {
                if (tbMaster.Visible)
                {
                    Value = (Single)(Value * 100.00 * (tbMaster.Value / 100.00));
                    pbMaster.Value = (int)Value;
                }
                else
                    pbMaster.Value = (int)(Value * 100);
            }

            if (FBars != null && FBars.Length > 0)
            {
                Single[] Values;
                Res = AudioMeter.GetChannelsPeak(out Values);
                if (Res == wclErrors.WCL_E_SUCCESS && Values != null && Values.Length > 0)
                {
                    for (Int32 i = 0; i < Values.Length; i++)
                        FBars[i].Value = (int)(Values[i] * 100);
                }
            }
        }

        private void Timer_Tick(Object sender, EventArgs e)
        {
            UpdateChannelsValue();
        }

        private void tbMaster_ValueChanged(Object sender, EventArgs e)
        {
            if (AudioVolume.Active)
            {
                Int32 Res = AudioVolume.SetVolume((float)(tbMaster.Value / 100.00));
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Trace("Set volume failed", Res);
            }
        }

        private void cbMute_CheckedChanged(Object sender, EventArgs e)
        {
            if (AudioVolume.Active)
            {
                Int32 Res = AudioVolume.SetMute(cbMute.Checked);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Trace("Set mute failed", Res);
            }
        }
    }
}
