using System;
using System.Windows.Forms;
using System.Text;

using wclCommon;
using wclBluetooth;
using wclCommunication;

namespace RfCommServer
{
    public partial class fmMain : Form
    {
        private wclRfCommServer Server;
        private wclBluetoothManager Manager;

        public fmMain()
        {
            InitializeComponent();
        }

        private void btClear_Click(object sender, EventArgs e)
        {
            lbEvents.Items.Clear();
        }

        private void fmMain_Load(object sender, EventArgs e)
        {
            Manager = new wclBluetoothManager();
            Server = new wclRfCommServer();

            Manager.OnNumericComparison += new wclBluetoothNumericComparisonEvent(Manager_OnNumericComparison);
            Manager.OnPasskeyNotification += new wclBluetoothPasskeyNotificationEvent(Manager_OnPasskeyNotification);
            Manager.OnPasskeyRequest += new wclBluetoothPasskeyRequestEvent(Manager_OnPasskeyRequest);
            Manager.OnPinRequest += new wclBluetoothPinRequestEvent(Manager_OnPinRequest);

            Server.OnClosed += new wclClientConnectionDisconnectEvent(Server_OnClosed);
            Server.OnConnect += new wclRfCommServerConnectEvent(Server_OnConnect);
            Server.OnData += new wclRfCommServerDataEvent(Server_OnData);
            Server.OnDisconnect += new wclRfCommServerDisconnectEvent(Server_OnDisconnect);
            Server.OnListen += new EventHandler(Server_OnListen);

            // In real application you should always analize the result code.
            // In this demo we assume that all is always OK.
            Manager.Open();
            
            cbAuthentication.Checked = Server.Authentication;
            cbEncryption.Checked = Server.Encryption;
        }

        void Server_OnListen(object sender, EventArgs e)
        {
            lbEvents.Items.Add("Server listening on radio " +
                ((wclRfCommServer)sender).Radio.ApiName + " on channel: " +
                ((wclRfCommServer)sender).AssignedChannel.ToString());
        }
        
        void  Server_OnDisconnect(object Sender, wclRfCommServerClientConnection Client, int Reason)
        {
            wclRfCommServerConnection Server = (wclRfCommServerConnection)Client.Server;
            
            lbEvents.Items.Add("Client disconnected: " + Client.Address.ToString("X12") + 
                " on radio " + Server.Radio.ApiName + " reason: 0x" + Reason.ToString("X8"));

            foreach (ListViewItem Item in lvClients.Items)
            {
                if (Item.Tag == Client)
                {
                    lvClients.Items.Remove(Item);
                    break;
                }
            }
        }
        
        void  Server_OnData(object Sender, wclRfCommServerClientConnection Client, byte[] Data)
        {
            if (Data != null && Data.Length > 0)
            {
                String Str = Encoding.ASCII.GetString(Data);

                wclRfCommServerConnection Server = (wclRfCommServerConnection)Client.Server;
                lbEvents.Items.Add("Received " + Data.Length.ToString() + " bytes: " + Str + " from " +
                    Client.Address.ToString("X12") + " on radio " + Server.Radio.ApiName);
            }
        }
        
        void  Server_OnConnect(object Sender, wclRfCommServerClientConnection Client, int Error)
        {
            wclRfCommServerConnection Server = (wclRfCommServerConnection)Client.Server;
            
            lbEvents.Items.Add("Client connected: " + Client.Address.ToString("X12") +
                " on Radio " + Server.Radio.ApiName + " error: 0x" + Error.ToString("X8"));

            ListViewItem Item = lvClients.Items.Add(Client.Address.ToString("X12"));
            Item.SubItems.Add(GetDeviceName(Client.Address));
            Item.Tag = Client;
        }

        void Server_OnClosed(object Sender, int Reason)
        {
            lbEvents.Items.Add("Server closed on radio " +
                ((wclRfCommServer)Sender).Radio.ApiName + "with reason: 0x" +Reason.ToString("X8"));
            
            lvClients.Items.Clear();
            
            edReadBuffer.Text = "";
            edWriteBuffer.Text = "";
        }

        void Manager_OnPinRequest(object Sender, wclBluetoothRadio Radio, long Address, out string Pin)
        {
            // Use '0000' as PIN.
            Pin = "0000";
            lbEvents.Items.Add("Pin request: " + Pin);
        }

        void Manager_OnPasskeyRequest(object Sender, wclBluetoothRadio Radio, long Address, out uint Passkey)
        {
            // Use 123456 as passkey.
            Passkey = 123456;
            lbEvents.Items.Add("Passkey request: " + Passkey.ToString());
        }

        void Manager_OnPasskeyNotification(object Sender, wclBluetoothRadio Radio, long Address, uint Passkey)
        {
            lbEvents.Items.Add("Passkey notification: " + Passkey.ToString());
        }

        void Manager_OnNumericComparison(object Sender, wclBluetoothRadio Radio, long Address, uint Number, out bool Confirm)
        {
            // Accept any pairing.
            Confirm = true;
            lbEvents.Items.Add("Numeric comparison: " + Number.ToString());
        }

        private void fmMain_FormClosed(object sender, FormClosedEventArgs e)
        {
            Server.Close();
            Server = null;

            Manager.Close();
            Manager = null;
        }

        private wclBluetoothRadio GetRadio()
        {
            wclBluetoothRadio Radio;
            Int32 Res = Manager.GetClassicRadio(out Radio);
            if (Res != wclErrors.WCL_E_SUCCESS)
            {
                MessageBox.Show("Get working radio failed: 0x" + Res.ToString("X8"));
                return null;
            }
            return Radio;
        }

        private void btListen_Click(object sender, EventArgs e)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                // Switch radio in discoverable and connectable mode.
                // Ignore any errors. In real application you need to check the result.
                Radio.SetDiscoverable(true);
                Radio.SetConnectable(true);

                Server.Authentication = cbAuthentication.Checked;
                Server.Encryption = cbEncryption.Checked;

                Server.Service = wclUUIDs.SerialPortServiceClass_UUID;
                Server.ServiceName = edServiceName.Text;

                Int32 Res = Server.Listen(Radio);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void btClose_Click(object sender, EventArgs e)
        {
            Int32 Res = Server.Close();
            if (Res != wclErrors.WCL_E_SUCCESS)
                MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        private void btDisconnect_Click(object sender, EventArgs e)
        {
            if (lvClients.SelectedItems.Count == 0)
                MessageBox.Show("Select client", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                Int32 Res = ((wclRfCommServerClientConnection)lvClients.SelectedItems[0].Tag).Disconnect();
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Error: 0x" + Res.ToString("X8"), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void btSend_Click(object sender, EventArgs e)
        {
            if (lvClients.SelectedItems.Count == 0)
                MessageBox.Show("Select client", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                Byte[] Ansi = Encoding.ASCII.GetBytes(edText.Text);
                UInt32 Sent = 0;
                Int32 Res = ((wclRfCommServerClientConnection)lvClients.SelectedItems[0].Tag).Write(Ansi, out Sent);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    MessageBox.Show("Error: 0x" + Res.ToString("X8"));
                else
                    MessageBox.Show("Sent: " + Sent.ToString() + " from " + Ansi.Length.ToString());
            }
        }

        private void btGetBuffers_Click(object sender, EventArgs e)
        {
            if (lvClients.SelectedItems.Count == 0)
                MessageBox.Show("Select client", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                wclRfCommServerClientConnection Client = (wclRfCommServerClientConnection)lvClients.SelectedItems[0].Tag;

                // Get read buffer.
                UInt32 Size = 0;
                Int32 Res = Client.GetReadBufferSize(out Size);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbEvents.Items.Add("Get read buffer error: 0x" + Res.ToString("X8"));
                else
                    edReadBuffer.Text = Size.ToString();

                // Get write buffer.
                Size = 0;
                Res = Client.GetWriteBufferSize(out Size);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbEvents.Items.Add("Get write buffer error: 0x" + Res.ToString("X8"));
                else
                    edWriteBuffer.Text = Size.ToString();
            }
        }

        private void btSetBuffers_Click(object sender, EventArgs e)
        {
            if (lvClients.SelectedItems.Count == 0)
                MessageBox.Show("Select client", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            else
            {
                wclRfCommServerClientConnection Client = (wclRfCommServerClientConnection)lvClients.SelectedItems[0].Tag;
                
                // Set read buffer.
                UInt32 Size = Convert.ToUInt32(edReadBuffer.Text);
                Int32 Res = Client.SetReadBufferSize(Size);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbEvents.Items.Add("Set read buffer error: 0x" + Res.ToString("X8"));
                
                // Set write buffer.
                Size = Convert.ToUInt32(edWriteBuffer.Text);
                Res = Client.SetWriteBufferSize(Size);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    lbEvents.Items.Add("Set write buffer error: 0x" + Res.ToString("X8"));
            }
        }

        private String GetDeviceName(Int64 Address)
        {
            wclBluetoothRadio Radio = GetRadio();
            if (Radio != null)
            {
                String DevName;
                Int32 Res = Radio.GetRemoteName(Address, out DevName);
                if (Res == wclErrors.WCL_E_SUCCESS)
                    return DevName;

                return "Error: 0x" + Res.ToString("X8");
            }

            return "";
        }
    }
}